from functools import partial
from typing import Optional, Union

import pandas as pd
import pandas_flavor as pf

from janitor.utils import check, check_column, deprecated_alias


@pf.register_dataframe_method
@deprecated_alias(col_name="column_name", type="cleaning_style")
def currency_column_to_numeric(
    df: pd.DataFrame,
    column_name: str,
    cleaning_style: Optional[str] = None,
    cast_non_numeric: Optional[dict] = None,
    fill_all_non_numeric: Optional[Union[float, int]] = None,
    remove_non_numeric: bool = False,
) -> pd.DataFrame:
    """Convert currency column to numeric.

    This method does not mutate the original DataFrame.

    This method allows one to take a column containing currency values,
    inadvertently imported as a string, and cast it as a float. This is
    usually the case when reading CSV files that were modified in Excel.
    Empty strings (i.e. `''`) are retained as `NaN` values.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "a_col": [" $24.56", "-", "($12.12)", "$1,000,000"],
        ...         "d_col": ["", "foo", "1.23 dollars", "-1,000 yen"],
        ...     }
        ... )
        >>> df  # doctest: +NORMALIZE_WHITESPACE
                a_col         d_col
        0      $24.56
        1           -           foo
        2   ($12.12)  1.23 dollars
        3  $1,000,000    -1,000 yen

        The default cleaning style.

        >>> df.currency_column_to_numeric("d_col")
                a_col    d_col
        0      $24.56      NaN
        1           -      NaN
        2   ($12.12)     1.23
        3  $1,000,000 -1000.00

        The accounting cleaning style.

        >>> df.currency_column_to_numeric(
        ...     "a_col", cleaning_style="accounting"
        ... )  # doctest: +NORMALIZE_WHITESPACE
                a_col         d_col
        0       24.56
        1        0.00           foo
        2      -12.12  1.23 dollars
        3  1000000.00    -1,000 yen

    Valid cleaning styles are:

    - `None`: Default cleaning is applied. Empty strings are always retained as
        `NaN`. Numbers, `-`, `.` are extracted and the resulting string
        is cast to a float.
    - `'accounting'`: Replaces numbers in parentheses with negatives, removes
        commas and currency symbols ($, €, £, ¥, ₹, ₩, ₽, ₱, ฿, ₺).

    Args:
        df: The pandas DataFrame.
        column_name: The column containing currency values to modify.
        cleaning_style: What style of cleaning to perform.
        cast_non_numeric: A dict of how to coerce certain strings to numeric
            type. For example, if there are values of 'REORDER' in the DataFrame,
            `{'REORDER': 0}` will cast all instances of 'REORDER' to 0.
            Only takes effect in the default cleaning style.
        fill_all_non_numeric: Similar to `cast_non_numeric`, but fills all
            strings to the same value. For example, `fill_all_non_numeric=1`, will
            make everything that doesn't coerce to a currency `1`.
            Only takes effect in the default cleaning style.
        remove_non_numeric: If set to True, rows of `df` that contain
            non-numeric values in the `column_name` column will be removed.
            Only takes effect in the default cleaning style.

    Raises:
        ValueError: If `cleaning_style` is not one of the accepted styles.

    Returns:
        A pandas DataFrame.
    """  # noqa: E501

    check("column_name", column_name, [str])
    check_column(df, column_name)

    column_series = df[column_name]
    if cleaning_style == "accounting":
        outcome = (
            df[column_name]
            .str.strip()
            .str.replace(r"[$€£¥₹₩₽₱฿₺]", "", regex=True)
            .str.replace(",", "", regex=False)
            .str.replace(")", "", regex=False)
            .str.replace("(", "-", regex=False)
            .replace({"-": 0.0})
            .astype(float)
        )
        return df.assign(**{column_name: outcome})
    if cleaning_style is not None:
        raise ValueError(
            "`cleaning_style` is expected to be one of ('accounting', None). "
            f"Got {cleaning_style!r} instead."
        )

    if cast_non_numeric:
        check("cast_non_numeric", cast_non_numeric, [dict])

    _make_cc_patrial = partial(
        _currency_column_to_numeric,
        cast_non_numeric=cast_non_numeric,
    )
    column_series = column_series.apply(_make_cc_patrial)

    if remove_non_numeric:
        df = df.loc[column_series != "", :]

    # _replace_empty_string_with_none is applied here after the check on
    # remove_non_numeric since "" is our indicator that a string was coerced
    # in the original column
    column_series = _replace_empty_string_with_none(column_series)

    if fill_all_non_numeric is not None:
        check("fill_all_non_numeric", fill_all_non_numeric, [int, float])
        column_series = column_series.fillna(fill_all_non_numeric)

    column_series = _replace_original_empty_string_with_none(column_series)

    df = df.assign(**{column_name: pd.to_numeric(column_series)})

    return df


def _currency_column_to_numeric(
    x: str,
    cast_non_numeric: Optional[dict] = None,
) -> Union[int, float, str]:
    """
    Perform logic for changing cell values.

    This is a private function intended to be used only in
    `currency_column_to_numeric`.

    It is intended to be used in a pandas `apply` method, after being passed
    through `partial`.

    Args:
        x: A string representing currency.
        cast_non_numeric: A dict of how to coerce certain strings to numeric
            type. For example, if there are values of 'REORDER' in the
            DataFrame, `{'REORDER': 0}` will cast all instances of 'REORDER'
            to 0.
    """
    acceptable_currency_characters = {
        "-",
        ".",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "0",
    }
    if not x:
        return "ORIGINAL_NA"

    if cast_non_numeric:
        if x in cast_non_numeric:
            mapped_x = cast_non_numeric[x]
            check(
                "{%r: %r}" % (x, str(mapped_x)),
                mapped_x,
                [int, float],
            )
            return mapped_x
    return "".join(i for i in x if i in acceptable_currency_characters)


def _replace_empty_string_with_none(column_series):
    column_series.loc[column_series == ""] = None
    return column_series


def _replace_original_empty_string_with_none(column_series):
    """Replaces original empty string with None"""
    column_series.loc[column_series == "ORIGINAL_NA"] = None
    return column_series



"""Time series-specific data cleaning functions."""

import itertools
from typing import Dict, Union

import pandas as pd
import pandas_flavor as pf

from .errors import JanitorError
from .utils import check


@pf.register_dataframe_method
def fill_missing_timestamps(
    df: pd.DataFrame,
    frequency: str,
    first_time_stamp: pd.Timestamp = None,
    last_time_stamp: pd.Timestamp = None,
) -> pd.DataFrame:
    """Fills a DataFrame with missing timestamps based on a defined frequency.

    If timestamps are missing, this function will re-index the DataFrame.
    If timestamps are not missing, then the function will return the DataFrame
    unmodified.

    Examples:
        Functional usage

        >>> import pandas as pd
        >>> import janitor.timeseries
        >>> df = janitor.timeseries.fill_missing_timestamps(
        ...     df=pd.DataFrame(...),
        ...     frequency="1h",
        ... )  # doctest: +SKIP

        Method chaining example:

        >>> import pandas as pd
        >>> import janitor.timeseries
        >>> df = pd.DataFrame(...).fill_missing_timestamps(
        ...     frequency="1h"
        ... )  # doctest: +SKIP

    Args:
        df: DataFrame which needs to be tested for missing timestamps
        frequency: Sampling frequency of the data.
            Acceptable frequency strings are available
            [here](https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases).
            Check offset aliases under time series in user guide
        first_time_stamp: Timestamp expected to start from;
            defaults to `None`. If no input is provided, assumes the
            minimum value in `time_series`.
        last_time_stamp: Timestamp expected to end with; defaults to `None`.
            If no input is provided, assumes the maximum value in
            `time_series`.

    Returns:
        DataFrame that has a complete set of contiguous datetimes.
    """
    # Check all the inputs are the correct data type
    check("frequency", frequency, [str])
    check("first_time_stamp", first_time_stamp, [pd.Timestamp, type(None)])
    check("last_time_stamp", last_time_stamp, [pd.Timestamp, type(None)])

    if first_time_stamp is None:
        first_time_stamp = df.index.min()
    if last_time_stamp is None:
        last_time_stamp = df.index.max()

    # Generate expected timestamps
    expected_timestamps = pd.date_range(
        start=first_time_stamp, end=last_time_stamp, freq=frequency
    )

    return df.reindex(expected_timestamps)


def _get_missing_timestamps(
    df: pd.DataFrame,
    frequency: str,
    first_time_stamp: pd.Timestamp = None,
    last_time_stamp: pd.Timestamp = None,
) -> pd.DataFrame:
    """Return the timestamps that are missing in a DataFrame.

    This function takes in a DataFrame, and checks its index
    against a DataFrame that contains the expected timestamps.
    Here, we assume that the expected timestamps are going to be
    of a larger size than the timestamps available in the input
    DataFrame `df`.

    If there are any missing timestamps in the input DataFrame,
    this function will return those missing timestamps from the
    expected DataFrame.
    """
    expected_df = df.fill_missing_timestamps(
        frequency, first_time_stamp, last_time_stamp
    )

    missing_timestamps = expected_df.index.difference(df.index)

    return expected_df.loc[missing_timestamps]


@pf.register_dataframe_method
def sort_timestamps_monotonically(
    df: pd.DataFrame, direction: str = "increasing", strict: bool = False
) -> pd.DataFrame:
    """Sort DataFrame such that index is monotonic.

    If timestamps are monotonic, this function will return
    the DataFrame unmodified. If timestamps are not monotonic,
    then the function will sort the DataFrame.

    Examples:
        Functional usage

        >>> import pandas as pd
        >>> import janitor.timeseries
        >>> df = janitor.timeseries.sort_timestamps_monotonically(
        ...     df=pd.DataFrame(...),
        ...     direction="increasing",
        ... )  # doctest: +SKIP

        Method chaining example:

        >>> import pandas as pd
        >>> import janitor.timeseries
        >>> df = pd.DataFrame(...).sort_timestamps_monotonically(
        ...     direction="increasing"
        ... )  # doctest: +SKIP

    Args:
        df: DataFrame which needs to be tested for monotonicity.
        direction: Type of monotonicity desired.
            Acceptable arguments are `'increasing'` or `'decreasing'`.
        strict: Flag to enable/disable strict monotonicity.
            If set to `True`, will remove duplicates in the index
            by retaining first occurrence of value in index.
            If set to `False`, will not test for duplicates in the index.

    Returns:
        DataFrame that has monotonically increasing (or decreasing)
            timestamps.
    """
    # Check all the inputs are the correct data type
    check("df", df, [pd.DataFrame])
    check("direction", direction, [str])
    check("strict", strict, [bool])

    # Remove duplicates if requested
    if strict:
        df = df[~df.index.duplicated(keep="first")]

    # Sort timestamps
    if direction == "increasing":
        df = df.sort_index()
    else:
        df = df.sort_index(ascending=False)

    # Return the DataFrame
    return df


def _flag_jumps_single_col(
    df: pd.DataFrame,
    col: str,
    scale: str,
    direction: str,
    threshold: Union[int, float],
) -> pd.Series:
    """
    Creates a boolean column that flags whether or not the change
    between consecutive rows in the provided DataFrame column exceeds a
    provided threshold.

    Comparisons are always performed utilizing a GREATER THAN
    threshold check. Thus, flags correspond to values that EXCEED
    the provided threshold.

    The method used to create consecutive row comparisons is set by the
    `scale` argument. A `scale=absolute` corresponds to a difference
    method (`.diff()`) and a `scale=percentage` corresponds to a
    percentage change methods (`pct_change()`).

    A `direction` argument is used to determine how to handle the sign
    of the difference or percentage change methods.
    A `direction=increasing` will only consider consecutive rows that
    are increasing in value and exceeding the provided threshold.
    A `direction=decreasing` will only consider consecutive rows that
    are decreasing in value and exceeding the provided threshold.
    If `direction=any`, the absolute value is taken for both the
    difference method and the percentage change methods and the sign
    between consecutive rows is ignored.
    """
    check("scale", scale, [str])
    check("direction", direction, [str])
    check("threshold", threshold, [int, float])

    scale_types = ["absolute", "percentage"]
    if scale not in scale_types:
        raise JanitorError(
            f"Unrecognized scale: '{scale}'. Must be one of: {scale_types}."
        )

    direction_types = ["increasing", "decreasing", "any"]
    if direction not in direction_types:
        raise JanitorError(
            f"Unrecognized direction: '{direction}'. Must be one of: {direction_types}."
        )

    if threshold < 0:
        raise JanitorError(
            f"Unrecognized threshold: {threshold}. "
            "This value must be >= 0.0. "
            "Use 'direction' to specify positive or negative intent."
        )

    single_col = df[col]
    single_col_diffs = single_col.diff()

    if scale == "percentage":
        single_col_pcts = single_col.pct_change()

        if direction == "increasing":
            # Using diffs ensures correct sign is used for incr/decr
            # (see issue #711)
            out = (single_col_diffs > 0) & (single_col_pcts.abs() > threshold)

        elif direction == "decreasing":
            # Using diffs ensures correct sign is used for incr/decr
            # (see issue #711)
            out = (single_col_diffs < 0) & (single_col_pcts.abs() > threshold)

        else:
            out = single_col_pcts.abs() > threshold

    else:
        if direction == "increasing":
            out = single_col_diffs > threshold

        elif direction == "decreasing":
            out = (single_col_diffs < 0) & (single_col_diffs.abs() > threshold)

        else:
            out = single_col_diffs.abs() > threshold

    out = out.astype(int)

    return out


@pf.register_dataframe_method
def flag_jumps(
    df: pd.DataFrame,
    scale: Union[str, Dict[str, str]] = "percentage",
    direction: Union[str, Dict[str, str]] = "any",
    threshold: Union[int, float, Dict[str, Union[int, float]]] = 0.0,
    strict: bool = False,
) -> pd.DataFrame:
    """Create boolean column(s) that flag whether or not the change
    between consecutive rows exceeds a provided threshold.

    Examples:

        Applies specified criteria across all columns of the DataFrame
        and appends a flag column for each column in the DataFrame

        >>> df = pd.DataFrame(...).flag_jumps(
        ...     scale="absolute", direction="any", threshold=2
        ... )  # doctest: +SKIP

        Applies specific criteria to certain DataFrame columns,
        applies default criteria to columns *not* specifically listed and
        appends a flag column for each column in the DataFrame

        >>> df = pd.DataFrame(...).flag_jumps(
        ...     scale=dict(col1="absolute", col2="percentage"),
        ...     direction=dict(col1="increasing", col2="any"),
        ...     threshold=dict(col1=1, col2=0.5),
        ... )  # doctest: +SKIP

        Applies specific criteria to certain DataFrame columns,
        applies default criteria to columns *not* specifically listed and
        appends a flag column for only those columns found in specified
        criteria

        >>> df = pd.DataFrame(...).flag_jumps(
        ...     scale=dict(col1="absolute"),
        ...     threshold=dict(col2=1),
        ...     strict=True,
        ... )  # doctest: +SKIP

    Args:
        df: DataFrame which needs to be flagged for changes between
            consecutive rows above a certain threshold.
        scale:
            Type of scaling approach to use.
            Acceptable arguments are

            * `'absolute'` (consider the difference between rows)
            * `'percentage'` (consider the percentage change between rows).

        direction: Type of method used to handle the sign change when
            comparing consecutive rows.
            Acceptable arguments are

            * `'increasing'` (only consider rows that are increasing in value)
            * `'decreasing'` (only consider rows that are decreasing in value)
            * `'any'` (consider rows that are either increasing or decreasing;
                sign is ignored).
        threshold: The value to check if consecutive row comparisons
            exceed. Always uses a greater than comparison. Must be `>= 0.0`.
        strict: Flag to enable/disable appending of a flag column for
            each column in the provided DataFrame. If set to `True`, will
            only append a flag column for those columns found in at least
            one of the input dictionaries. If set to `False`, will append
            a flag column for each column found in the provided DataFrame.
            If criteria is not specified, the defaults for each criteria
            is used.

    Raises:
        JanitorError: If `strict=True` and at least one of
            `scale`, `direction`, or `threshold` inputs is not a
            dictionary.
        JanitorError: If `scale` is not one of
            `("absolute", "percentage")`.
        JanitorError: If `direction` is not one of
            `("increasing", "decreasing", "any")`.
        JanitorError: If `threshold` is less than `0.0`.

    Returns:
        DataFrame that has `flag jump` columns.

    <!--
    # noqa: DAR101
    -->
    """
    df = df.copy()

    if strict:
        if any(isinstance(arg, dict) for arg in (scale, direction, threshold)) is False:
            raise JanitorError(
                "When enacting 'strict=True', 'scale', 'direction', or "
                + "'threshold' must be a dictionary."
            )

        # Only append a flag col for the cols that appear
        # in at least one of the input dicts
        arg_keys = [
            arg.keys() for arg in (scale, direction, threshold) if isinstance(arg, dict)
        ]
        cols = set(itertools.chain.from_iterable(arg_keys))

    else:
        # Append a flag col for each col in the DataFrame
        cols = df.columns

    columns_to_add = {}
    for col in sorted(cols):
        # Allow arguments to be a mix of dict and single instances
        s = scale.get(col, "percentage") if isinstance(scale, dict) else scale
        d = direction.get(col, "any") if isinstance(direction, dict) else direction
        t = threshold.get(col, 0.0) if isinstance(threshold, dict) else threshold

        columns_to_add[f"{col}_jump_flag"] = _flag_jumps_single_col(
            df, col, scale=s, direction=d, threshold=t
        )

    df = df.assign(**columns_to_add)

    return df



"""Utility functions for all of the functions submodule."""

from __future__ import annotations

import re
import unicodedata
import warnings
from enum import Enum
from typing import (
    Any,
    Hashable,
    Iterable,
    List,
    Optional,
    Pattern,
    Union,
)

import numpy as np
import pandas as pd
from multipledispatch import dispatch
from pandas.api.types import (
    is_list_like,
    is_scalar,
    is_string_dtype,
    union_categoricals,
)

from janitor.errors import JanitorError
from janitor.utils import (
    _expand_grid,
    check,
    check_column,
    find_stack_level,
)

warnings.simplefilter("always", DeprecationWarning)


def unionize_dataframe_categories(
    *dataframes: Any,
    column_names: Optional[Iterable[pd.CategoricalDtype]] = None,
) -> List[pd.DataFrame]:
    """
    Given a group of dataframes which contain some categorical columns, for
    each categorical column present, find all the possible categories across
    all the dataframes which have that column.
    Update each dataframes' corresponding column with a new categorical object
    that contains the original data
    but has labels for all the possible categories from all dataframes.
    This is useful when concatenating a list of dataframes which all have the
    same categorical columns into one dataframe.

    If, for a given categorical column, all input dataframes do not have at
    least one instance of all the possible categories,
    Pandas will change the output dtype of that column from `category` to
    `object`, losing out on dramatic speed gains you get from the former
    format.

    Examples:
        Usage example for concatenation of categorical column-containing
        dataframes:

        Instead of:

        ```python
        concatenated_df = pd.concat([df1, df2, df3], ignore_index=True)
        ```

        which in your case has resulted in `category` -> `object` conversion,
        use:

        ```python
        unionized_dataframes = unionize_dataframe_categories(df1, df2, df2)
        concatenated_df = pd.concat(unionized_dataframes, ignore_index=True)
        ```

    Args:
        *dataframes: The dataframes you wish to unionize the categorical
            objects for.
        column_names: If supplied, only unionize this subset of columns.

    Raises:
        TypeError: If any of the inputs are not pandas DataFrames.

    Returns:
        A list of the category-unioned dataframes in the same order they
            were provided.
    """

    if any(not isinstance(df, pd.DataFrame) for df in dataframes):
        raise TypeError("Inputs must all be dataframes.")

    if column_names is None:
        # Find all columns across all dataframes that are categorical

        column_names = set()

        for dataframe in dataframes:
            column_names = column_names.union(
                [
                    column_name
                    for column_name in dataframe.columns
                    if isinstance(dataframe[column_name].dtype, pd.CategoricalDtype)
                ]
            )

    else:
        column_names = [column_names]
    # For each categorical column, find all possible values across the DFs

    category_unions = {
        column_name: union_categoricals(
            [df[column_name] for df in dataframes if column_name in df.columns]
        )
        for column_name in column_names
    }

    # Make a shallow copy of all DFs and modify the categorical columns
    # such that they can encode the union of all possible categories for each.

    refactored_dfs = []

    for df in dataframes:
        df = df.copy(deep=False)

        for column_name, categorical in category_unions.items():
            if column_name in df.columns:
                df[column_name] = pd.Categorical(
                    df[column_name], categories=categorical.categories
                )

        refactored_dfs.append(df)

    return refactored_dfs


def patterns(regex_pattern: Union[str, Pattern]) -> Pattern:
    """This function converts a string into a compiled regular expression.

    It can be used to select columns in the index or columns_names
    arguments of `pivot_longer` function.

    !!!warning

        This function is deprecated. Kindly use `re.compile` instead.

    Args:
        regex_pattern: String to be converted to compiled regular
            expression.

    Returns:
        A compile regular expression from provided `regex_pattern`.
    """
    warnings.warn(
        "This function is deprecated. Kindly use `re.compile` instead.",
        DeprecationWarning,
        stacklevel=find_stack_level(),
    )
    check("regular expression", regex_pattern, [str, Pattern])

    return re.compile(regex_pattern)


def _computations_expand_grid(others: dict) -> dict:
    """
    Creates a cartesian product of all the inputs in `others`.
    Uses numpy's `mgrid` to generate indices, which is used to
    `explode` all the inputs in `others`.

    There is a performance penalty for small entries
    in using this method, instead of `itertools.product`;
    however, there are significant performance benefits
    as the size of the data increases.

    Another benefit of this approach, in addition to the significant
    performance gains, is the preservation of data types.
    This is particularly relevant for pandas' extension arrays `dtypes`
    (categoricals, nullable integers, ...).

    A dictionary of all possible combinations is returned.
    """

    for key in others:
        check("key", key, [Hashable])

    grid = {}

    for key, value in others.items():
        if is_scalar(value):
            value = np.asarray([value])
        elif is_list_like(value) and (not hasattr(value, "shape")):
            value = np.asarray([*value])
        if not value.size:
            raise ValueError(f"Kindly provide a non-empty array for {key}.")

        grid[key] = value

    others = None

    # slice obtained here is used in `np.mgrid`
    # to generate cartesian indices
    # which is then paired with grid.items()
    # to blow up each individual value
    # before creating the final DataFrame.
    grid = grid.items()
    grid_index = [slice(len(value)) for _, value in grid]
    grid_index = map(np.ravel, np.mgrid[grid_index])
    grid = zip(grid, grid_index)
    grid = ((*left, right) for left, right in grid)
    contents = {}
    for key, value, grid_index in grid:
        contents.update(_expand_grid(value, grid_index, key))
    # check length of keys and pad if necessary
    lengths = set(map(len, contents))
    if len(lengths) > 1:
        lengths = max(lengths)
        others = {}
        for key, value in contents.items():
            len_key = len(key)
            if len_key < lengths:
                padding = [""] * (lengths - len_key)
                key = (*key, *padding)
            others[key] = value
        return others
    return contents


@dispatch(pd.DataFrame, (list, tuple), str)
def _factorize(df, column_names, suffix, **kwargs):
    check_column(df, column_names=column_names, present=True)
    for col in column_names:
        df[f"{col}{suffix}"] = pd.factorize(df[col], **kwargs)[0]
    return df


@dispatch(pd.DataFrame, str, str)
def _factorize(df, column_name, suffix, **kwargs):  # noqa: F811
    check_column(df, column_names=column_name, present=True)
    df[f"{column_name}{suffix}"] = pd.factorize(df[column_name], **kwargs)[0]
    return df


class _JoinOperator(Enum):
    """
    List of operators used in conditional_join.
    """

    GREATER_THAN = ">"
    LESS_THAN = "<"
    GREATER_THAN_OR_EQUAL = ">="
    LESS_THAN_OR_EQUAL = "<="
    STRICTLY_EQUAL = "=="
    NOT_EQUAL = "!="


less_than_join_types = {
    _JoinOperator.LESS_THAN.value,
    _JoinOperator.LESS_THAN_OR_EQUAL.value,
}
greater_than_join_types = {
    _JoinOperator.GREATER_THAN.value,
    _JoinOperator.GREATER_THAN_OR_EQUAL.value,
}


def _null_checks_cond_join(left: pd.Series, right: pd.Series) -> Union[tuple, None]:
    """
    Checks for nulls in the arrays before conducting binary search.

    Relevant to _less_than_indices and _greater_than_indices
    """
    any_nulls = left.isna()
    if any_nulls.all():
        return None
    if any_nulls.any():
        left = left[~any_nulls]
    any_nulls = right.isna()
    if any_nulls.all():
        return None
    if any_nulls.any():
        right = right[~any_nulls]
    any_nulls = any_nulls.any()
    right_is_sorted = right.is_monotonic_increasing
    if not right_is_sorted:
        right = right.sort_values(kind="stable")

    left_index = left.index._values
    left = left._values
    right_index = right.index._values
    right = right._values

    return left, right, left_index, right_index, right_is_sorted, any_nulls


def _equal_indices(
    left: pd.Series,
    right: pd.Series,
    return_ragged_arrays: bool,
    row_count: Hashable = None,
) -> tuple:
    """
    Use binary search to get indices where left
    is equal to right.

    A tuple of integer indexes
    for left and right is returned.
    """
    outcome = _null_checks_cond_join(left=left, right=right)
    if not outcome:
        return None
    left, right, left_index, right_index, right_is_sorted, any_nulls = outcome
    # steal some perf here within the binary search
    # search for uniques
    # and later index them with left_positions
    # it is assumed that users will only reach for this
    # if the data is reasonably duplicated; if not
    # pd.merge is superb especially if it's a one-to-one
    # or one-to-many
    positions, left = pd.factorize(left, sort=False)
    if return_ragged_arrays:
        starts = right.searchsorted(left, side="left")
        starts = starts[positions]
        ends = right.searchsorted(left, side="right")
        ends = ends[positions]
        booleans = starts < ends
        if not booleans.any():
            return None
        if not booleans.all():
            left_index = left_index[booleans]
            starts = starts[booleans]
            ends = ends[booleans]
        right = [slice(start, end) for start, end in zip(starts, ends)]
        if right_is_sorted & (not any_nulls):
            return left_index, right
        right = [right_index[slicer] for slicer in right]
        return left_index, right
    # necessary step to remove non matches in right
    # vital to ensuring correct output in numba_equi_join
    # when building the regions
    booleans = pd.Index(left).get_indexer(right) != -1
    if not booleans.any():
        return None
    if not booleans.all():
        right_index = right_index[booleans]
        right = right[booleans]
    starts = right.searchsorted(left, side="left")
    starts = starts[positions]
    ends = right.searchsorted(left, side="right")
    ends = ends[positions]
    booleans = starts < ends
    if not booleans.any():
        return None
    if not booleans.all():
        left_index = left_index[booleans]
        starts = starts[booleans]
    if row_count:
        return pd.Series(index=left_index, data=ends - starts, name=row_count)
    return left_index, right_index, starts


def _less_than_indices(
    left: pd.Series,
    right: pd.Series,
    strict: bool,
    multiple_conditions: bool,
    keep: str,
    return_ragged_arrays: bool,
    row_count: Hashable = None,
) -> tuple:
    """
    Use binary search to get indices where left
    is less than or equal to right.

    If strict is True, then only indices
    where `left` is less than
    (but not equal to) `right` are returned.

    A tuple of integer indexes
    for left and right is returned.
    """

    # no point going through all the hassle
    if left.min() > right.max():
        return None

    outcome = _null_checks_cond_join(left=left, right=right)
    if not outcome:
        return None
    left, right, left_index, right_index, right_is_sorted, any_nulls = outcome

    search_indices = right.searchsorted(left, side="left")
    # if any of the positions in `search_indices`
    # is equal to the length of `right_keys`
    # that means the respective position in `left`
    # has no values from `right` that are less than
    # or equal, and should therefore be discarded
    len_right = right.size
    booleans = search_indices < len_right

    if not booleans.all():
        left = left[booleans]
        left_index = left_index[booleans]
        search_indices = search_indices[booleans]

    # the idea here is that if there are any equal values
    # shift to the right to the immediate next position
    # that is not equal
    if strict:
        booleans = left == right[search_indices]
        # replace positions where rows are equal
        # with positions from searchsorted('right')
        # positions from searchsorted('right') will never
        # be equal and will be the furthermost in terms of position
        # example : right -> [2, 2, 2, 3], and we need
        # positions where values are not equal for 2;
        # the furthermost will be 3, and searchsorted('right')
        # will return position 3.
        if booleans.any():
            replacements = right.searchsorted(left, side="right")
            # now we can safely replace values
            # with strictly less than positions
            search_indices = np.where(booleans, replacements, search_indices)
        # check again if any of the values
        # have become equal to length of right
        # and get rid of them
        booleans = search_indices < len_right

        if not booleans.all():
            left_index = left_index[booleans]
            search_indices = search_indices[booleans]

        if not search_indices.size:
            return None
    if row_count:
        return pd.Series(
            index=left_index, data=len_right - search_indices, name=row_count
        )
    if multiple_conditions:
        return left_index, right_index, search_indices
    if right_is_sorted & (keep == "last"):
        indexer = np.empty_like(search_indices)
        indexer[:] = len_right - 1
        return left_index, right_index[indexer]
    if right_is_sorted & (keep == "first") & any_nulls:
        return left_index, right_index[search_indices]
    if right_is_sorted & (keep == "first"):
        return left_index, search_indices
    if return_ragged_arrays & right_is_sorted & (not any_nulls):
        right = [slice(ind, len_right) for ind in search_indices]
        return left_index, right
    right = [right_index[ind:len_right] for ind in search_indices]
    if return_ragged_arrays:
        return left_index, right
    if keep == "first":
        right = [arr.min() for arr in right]
        return left_index, right
    if keep == "last":
        right = [arr.max() for arr in right]
        return left_index, right
    right = np.concatenate(right)
    left = left_index.repeat(len_right - search_indices)
    return left, right


def _greater_than_indices(
    left: pd.Series,
    right: pd.Series,
    strict: bool,
    multiple_conditions: bool,
    keep: str,
    return_ragged_arrays: bool,
    row_count: Hashable = None,
) -> tuple:
    """
    Use binary search to get indices where left
    is greater than or equal to right.

    If strict is True, then only indices
    where `left` is greater than
    (but not equal to) `right` are returned.

    if multiple_conditions is False, a tuple of integer indexes
    for left and right is returned;
    else a tuple of the index for left, right, as well
    as the positions of left in right is returned.
    """

    # quick break, avoiding the hassle
    if left.max() < right.min():
        return None

    outcome = _null_checks_cond_join(left=left, right=right)
    if not outcome:
        return None
    left, right, left_index, right_index, right_is_sorted, any_nulls = outcome
    search_indices = right.searchsorted(left, side="right")
    # if any of the positions in `search_indices`
    # is equal to 0 (less than 1), it implies that
    # left[position] is not greater than any value
    # in right
    booleans = search_indices > 0
    if not booleans.all():
        left = left[booleans]
        left_index = left_index[booleans]
        search_indices = search_indices[booleans]

    # the idea here is that if there are any equal values
    # shift downwards to the immediate next position
    # that is not equal
    if strict:
        booleans = left == right[search_indices - 1]
        # replace positions where rows are equal with
        # searchsorted('left');
        # this works fine since we will be using the value
        # as the right side of a slice, which is not included
        # in the final computed value
        if booleans.any():
            replacements = right.searchsorted(left, side="left")
            # now we can safely replace values
            # with strictly greater than positions
            search_indices = np.where(booleans, replacements, search_indices)
        # any value less than 1 should be discarded
        # since the lowest value for binary search
        # with side='right' should be 1
        booleans = search_indices > 0
        if not booleans.all():
            left_index = left_index[booleans]
            search_indices = search_indices[booleans]

        if not search_indices.size:
            return None
    if row_count:
        return pd.Series(index=left_index, data=search_indices, name=row_count)
    if multiple_conditions:
        return left_index, right_index, search_indices
    if right_is_sorted & (keep == "first"):
        indexer = np.zeros_like(search_indices)
        return left_index, right_index[indexer]
    if right_is_sorted & (keep == "last") & any_nulls:
        return left_index, right_index[search_indices - 1]
    if right_is_sorted & (keep == "last"):
        return left_index, search_indices - 1
    if return_ragged_arrays & right_is_sorted & (not any_nulls):
        right = [slice(0, ind) for ind in search_indices]
        return left_index, right
    right = [right_index[:ind] for ind in search_indices]
    if return_ragged_arrays:
        return left_index, right
    if keep == "first":
        right = [arr.min() for arr in right]
        return left_index, right
    if keep == "last":
        right = [arr.max() for arr in right]
        return left_index, right
    right = np.concatenate(right)
    left = left_index.repeat(search_indices)
    return left, right


def _not_equal_indices(left: pd.Series, right: pd.Series, keep: str) -> tuple:
    """
    Use binary search to get indices where
    `left` is exactly  not equal to `right`.

    It is a combination of strictly less than
    and strictly greater than indices.

    A tuple of integer indexes for left and right
    is returned.
    """

    dummy = np.array([], dtype=int)

    # deal with nulls
    l1_nulls = dummy
    r1_nulls = dummy
    l2_nulls = dummy
    r2_nulls = dummy
    any_left_nulls = left.isna()
    any_right_nulls = right.isna()
    if any_left_nulls.any():
        l1_nulls = left.index[any_left_nulls.array]
        l1_nulls = l1_nulls.to_numpy(copy=False)
        r1_nulls = right.index
        # avoid NAN duplicates
        if any_right_nulls.any():
            r1_nulls = r1_nulls[~any_right_nulls.array]
        r1_nulls = r1_nulls.to_numpy(copy=False)
        nulls_count = l1_nulls.size
        # blow up nulls to match length of right
        l1_nulls = np.tile(l1_nulls, r1_nulls.size)
        # ensure length of right matches left
        if nulls_count > 1:
            r1_nulls = np.repeat(r1_nulls, nulls_count)
    if any_right_nulls.any():
        r2_nulls = right.index[any_right_nulls.array]
        r2_nulls = r2_nulls.to_numpy(copy=False)
        l2_nulls = left.index
        nulls_count = r2_nulls.size
        # blow up nulls to match length of left
        r2_nulls = np.tile(r2_nulls, l2_nulls.size)
        # ensure length of left matches right
        if nulls_count > 1:
            l2_nulls = np.repeat(l2_nulls, nulls_count)

    l1_nulls = np.concatenate([l1_nulls, l2_nulls])
    r1_nulls = np.concatenate([r1_nulls, r2_nulls])

    outcome = _less_than_indices(
        left,
        right,
        strict=True,
        multiple_conditions=False,
        keep=keep,
        return_ragged_arrays=False,
    )

    if outcome is None:
        lt_left = dummy
        lt_right = dummy
    else:
        lt_left, lt_right = outcome

    outcome = _greater_than_indices(
        left,
        right,
        strict=True,
        multiple_conditions=False,
        keep=keep,
        return_ragged_arrays=False,
    )

    if outcome is None:
        gt_left = dummy
        gt_right = dummy
    else:
        gt_left, gt_right = outcome

    left = np.concatenate([lt_left, gt_left, l1_nulls])
    right = np.concatenate([lt_right, gt_right, r1_nulls])

    if (not left.size) & (not right.size):
        return None
    return _keep_output(keep, left, right)


def _not_equal_row_count(
    left: pd.Series, right: pd.Series, row_count: Hashable
) -> tuple | None:
    """
    Get row count where
    `left` is exactly  not equal to `right`.
    """

    null_count = pd.Series(index=left.index, data=right.isna().sum(), name=row_count)
    null_count[left.isna()] += right.notna().sum()
    outcome = _less_than_indices(
        left,
        right,
        strict=True,
        multiple_conditions=False,
        keep="all",
        return_ragged_arrays=False,
        row_count=row_count,
    )
    if outcome is not None:
        null_count = null_count.add(outcome, fill_value=0)
    outcome = _greater_than_indices(
        left,
        right,
        strict=True,
        multiple_conditions=False,
        keep="all",
        return_ragged_arrays=False,
        row_count=row_count,
    )
    if outcome is not None:
        null_count = null_count.add(outcome, fill_value=0)
    if not null_count.sum(axis=None):
        return None
    return null_count


def _generic_func_cond_join(
    left: pd.Series,
    right: pd.Series,
    op: str,
    multiple_conditions: bool,
    keep: str,
    row_count: Hashable = None,
    return_ragged_arrays: bool = False,
) -> tuple:
    """
    Generic function to call any of the individual functions
    (_less_than_indices, _greater_than_indices,
    or _not_equal_indices).
    """
    strict = False

    if op in {
        _JoinOperator.GREATER_THAN.value,
        _JoinOperator.LESS_THAN.value,
        _JoinOperator.NOT_EQUAL.value,
    }:
        strict = True

    if op in less_than_join_types:
        return _less_than_indices(
            left=left,
            right=right,
            strict=strict,
            multiple_conditions=multiple_conditions,
            keep=keep,
            return_ragged_arrays=return_ragged_arrays,
            row_count=row_count,
        )
    if op in greater_than_join_types:
        return _greater_than_indices(
            left=left,
            right=right,
            strict=strict,
            multiple_conditions=multiple_conditions,
            keep=keep,
            return_ragged_arrays=return_ragged_arrays,
            row_count=row_count,
        )
    if (op == _JoinOperator.NOT_EQUAL.value) and row_count:
        return _not_equal_row_count(left=left, right=right, row_count=row_count)
    if op == _JoinOperator.NOT_EQUAL.value:
        return _not_equal_indices(left=left, right=right, keep=keep)
    return _equal_indices(
        left=left,
        right=right,
        return_ragged_arrays=return_ragged_arrays,
        row_count=row_count,
    )


def _keep_output(keep: str, left: np.ndarray, right: np.ndarray):
    """return indices for left and right index based on the value of `keep`."""
    if keep == "all":
        return left, right
    grouped = pd.Series(right).groupby(left)
    if keep == "first":
        grouped = grouped.min()
        return grouped.index, grouped._values
    grouped = grouped.max()
    return grouped.index, grouped._values


def _change_case(
    obj: str,
    case_type: str,
) -> str:
    """Change case of obj."""
    case_types = {"preserve", "upper", "lower", "snake"}
    case_type = case_type.lower()
    if case_type not in case_types:
        raise JanitorError(f"type must be one of: {case_types}")

    if case_type == "preserve":
        return obj
    if case_type == "upper":
        return obj.upper()
    if case_type == "lower":
        return obj.lower()
    # Implementation adapted from: https://gist.github.com/jaytaylor/3660565
    # by @jtaylor
    obj = re.sub(pattern=r"(.)([A-Z][a-z]+)", repl=r"\1_\2", string=obj)
    obj = re.sub(pattern=r"([a-z0-9])([A-Z])", repl=r"\1_\2", string=obj)
    return obj.lower()


def _normalize_1(obj: str) -> str:
    """Perform normalization of obj."""
    FIXES = [
        (r"[ /:,?()\.-]", "_"),
        (r"['‘’]", ""),
        (r"[\xa0]", "_"),
        # Use capturing groups instead of lookbehind/lookahead
        # for compatibility with pyarrow's RE2 engine (pandas 3.0+).
        (r"(\w)@(\w)", r"\1_\2"),
    ]
    for search, replace in FIXES:
        obj = re.sub(pattern=search, repl=replace, string=obj)

    return obj


def _remove_special(
    obj: str,
) -> str:
    """Remove special characters from obj."""
    obj = [item for item in obj if item.isalnum() or (item == "_")]
    return "".join(obj)


def _strip_accents(
    obj: str,
) -> str:
    """Remove accents from obj.

    Inspired from [StackOverflow][so].

    [so]: https://stackoverflow.com/questions/517923/what-is-the-best-way-to-remove-accents-in-a-python-unicode-strin
    """  # noqa: E501

    obj = [
        letter
        for letter in unicodedata.normalize("NFD", obj)
        if not unicodedata.combining(letter)
    ]
    return "".join(obj)


def _strip_underscores_func(
    obj: str,
    strip_underscores: Union[str, bool] = None,
) -> str:
    """Strip underscores from obj."""
    underscore_options = {None, "left", "right", "both", "l", "r", True}
    if strip_underscores not in underscore_options:
        raise JanitorError(f"strip_underscores must be one of: {underscore_options}")

    if strip_underscores in {"left", "l"}:
        return obj.lstrip("_")
    if strip_underscores in {"right", "r"}:
        return obj.rstrip("_")
    if strip_underscores in {True, "both"}:
        return obj.strip("_")
    return obj


def _is_str_or_cat(index):
    """
    Check if the column/index is a string,
    or categorical with strings.
    """
    if isinstance(index.dtype, pd.CategoricalDtype):
        return is_string_dtype(index.categories)
    return is_string_dtype(index)



from __future__ import annotations

import janitor_rs
import numpy as np
import pandas as pd

from janitor.functions._conditional_join import (
    _equi_helpers,
    _range_indices,
)


def _get_indices(
    df: pd.DataFrame,
    right: pd.DataFrame,
    mapping: dict,
    return_matching_indices: bool,
    keep: str,
) -> dict:
    """Get indices for an equi join only"""
    # purely equi join, and return_matching_indices is True
    l_cols, r_cols = _equi_helpers._build_equi_indices(
        df=df, right=right, mapping=mapping
    )
    positions, uniques, starts, ends = _equi_helpers._get_positions_right(
        right_columns=r_cols
    )
    check = r_cols.is_monotonic_increasing
    if not check:
        reordered_positions = janitor_rs.reorder_index(
            positions=positions, starts=starts
        )
        right = right.iloc[reordered_positions]
    outcome = _equi_helpers._get_indexers(
        l_cols=l_cols, uniques=uniques, starts=starts, ends=ends
    )
    if outcome is None:
        return {
            "left_index": np.array([], dtype=np.intp),
            "right_index": np.array([], dtype=np.intp),
        }
    indexers, starts, ends = outcome
    booleans = indexers == -1
    if booleans.all():
        return {
            "left_index": np.array([], dtype=np.intp),
            "right_index": np.array([], dtype=np.intp),
        }
    left_index = df.index._values
    if booleans.any():
        booleans = ~booleans
        starts = starts[booleans]
        ends = ends[booleans]
        left_index = left_index[booleans]
    right_index = right.index._values
    if return_matching_indices:
        return {
            "left_index": left_index,
            "right_index": right_index,
            "starts": starts,
            "ends": ends,
        }
    return _range_indices._build_indices(
        left_index=left_index,
        right_index=right_index,
        starts=starts,
        ends=ends,
        keep=keep,
        right_is_sorted=check,
    )



"""Miscellaneous utility functions."""

from __future__ import annotations

import importlib
import os
import socket
import sys
from functools import singledispatch, wraps
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Union
from warnings import warn

import numpy as np
import pandas as pd


def check(varname: str, value, expected_types: list):
    """One-liner syntactic sugar for checking types.

    It can also check callables.

    Examples:
        ```python
        check("x", x, [int, float])
        ```

    Args:
        varname: The name of the variable (for diagnostic error message).
        value: The value of the `varname`.
        expected_types: The type(s) the item is expected to be.

    Raises:
        TypeError: If data is not the expected type.
    """
    is_expected_type: bool = False
    for t in expected_types:
        if t is callable:
            is_expected_type = t(value)
        else:
            is_expected_type = isinstance(value, t)
        if is_expected_type:
            break

    if not is_expected_type:
        raise TypeError(f"{varname} should be one of {expected_types}.")


@singledispatch
def _expand_grid(value, grid_index, key):
    """
    Base function for dispatch of `_expand_grid`.
    """

    raise TypeError(
        f"{type(value).__name__} data type is not supported in `expand_grid`."
    )


@_expand_grid.register(np.ndarray)
def _sub_expand_grid(value, grid_index, key):  # noqa: F811
    """
    Expands the numpy array based on `grid_index`.
    Returns a dictionary.
    """

    if value.ndim > 2:
        raise ValueError(
            "expand_grid works only on 1D and 2D arrays. "
            f"The provided array for {key} however "
            f"has a dimension of {value.ndim}."
        )

    value = value[grid_index]

    if value.ndim == 1:
        return {(key, 0): value}

    return {(key, num): value[:, num] for num in range(value.shape[-1])}


@_expand_grid.register(pd.api.extensions.ExtensionArray)
def _sub_expand_grid(value, grid_index, key):  # noqa: F811
    """
    Expands the pandas array based on `grid_index`.
    Returns a dictionary.
    """

    return {(key, 0): value[grid_index]}


@_expand_grid.register(pd.Index)
@_expand_grid.register(pd.Series)
def _sub_expand_grid(value, grid_index, key):  # noqa: F811
    """
    Expands the pd.Series/pd.Index based on `grid_index`.
    Returns a dictionary.
    """

    name = value.name or 0

    return {(key, name): value._values[grid_index]}


@_expand_grid.register(pd.DataFrame)
def _sub_expand_grid(value, grid_index, key):  # noqa: F811
    """
    Expands the DataFrame based on `grid_index`.
    Returns a dictionary.
    """
    if isinstance(value.columns, pd.MultiIndex):
        return {(key, *name): val._values[grid_index] for name, val in value.items()}

    return {(key, name): val._values[grid_index] for name, val in value.items()}


@_expand_grid.register(pd.MultiIndex)
def _sub_expand_grid(value, grid_index, key):  # noqa: F811
    """
    Expands the MultiIndex based on `grid_index`.
    Returns a dictionary.
    """

    contents = {}
    num = 0
    for n in range(value.nlevels):
        arr = value.get_level_values(n)
        name = arr.name
        arr = arr._values[grid_index]
        if not name:
            name = num
            num += 1
        contents[(key, name)] = arr
    return contents


def import_message(
    submodule: str,
    package: str,
    conda_channel: str = None,
    pip_install: bool = False,
):
    """Return warning if package is not found.

    Generic message for indicating to the user when a function relies on an
    optional module / package that is not currently installed. Includes
    installation instructions. Used in `chemistry.py` and `biology.py`.

    Args:
        submodule: `pyjanitor` submodule that needs an external dependency.
        package: External package this submodule relies on.
        conda_channel: `conda` channel package can be installed from,
            if at all.
        pip_install: Whether package can be installed via `pip`.
    """
    is_conda = os.path.exists(os.path.join(sys.prefix, "conda-meta"))
    installable = True
    if is_conda:
        if conda_channel is None:
            installable = False
            installation = f"{package} cannot be installed via conda"
        else:
            installation = f"conda install -c {conda_channel} {package}"
    else:
        if pip_install:
            installation = f"pip install {package}"
        else:
            installable = False
            installation = f"{package} cannot be installed via pip"

    print(f"To use the janitor submodule {submodule}, you need to install {package}.")
    print()
    if installable:
        print("To do so, use the following command:")
        print()
        print(f"    {installation}")
    else:
        print(f"{installation}")


def idempotent(func: Callable, df: pd.DataFrame, *args: Any, **kwargs: Any):
    """Raises an error if a function operating on a DataFrame is not
    idempotent.

    That is, `func(func(df)) = func(df)` is not `True` for all `df`.

    Args:
        func: A Python method.
        df: A pandas `DataFrame`.
        *args: Positional arguments supplied to the method.
        **kwargs: Keyword arguments supplied to the method.

    Raises:
        ValueError: If `func` is found to not be idempotent for the given
            DataFrame (`df`).
    """
    if not func(df, *args, **kwargs) == func(
        func(df, *args, **kwargs), *args, **kwargs
    ):
        raise ValueError("Supplied function is not idempotent for the given DataFrame.")


def deprecated_kwargs(
    *arguments: List[str],
    message: str = (
        "The keyword argument '{argument}' of '{func_name}' is deprecated."
    ),
    error: bool = True,
) -> Callable:
    """Used as a decorator when deprecating function's keyword arguments.

    Examples:

        ```python
        from janitor.utils import deprecated_kwargs


        @deprecated_kwargs("x", "y")
        def plus(a, b, x=0, y=0):
            return a + b
        ```

    Args:
        *arguments: The list of deprecated keyword arguments.
        message: The message of `ValueError` or `DeprecationWarning`.
            It should be a string or a string template. If a string template
            defaults input `func_name` and `argument`.
        error: If True, raises `ValueError` else returns `DeprecationWarning`.

    Raises:
        ValueError: If one of `arguments` is in the decorated function's
            keyword arguments.

    Returns:
        The original function wrapped with the deprecated `kwargs`
            checking function.

    <!--
    # noqa: DAR402
    -->
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for argument in arguments:
                if argument in kwargs:
                    msg = message.format(
                        func_name=func.__name__,
                        argument=argument,
                    )
                    if error:
                        raise ValueError(msg)
                    else:
                        warn(msg, DeprecationWarning)

            return func(*args, **kwargs)

        return wrapper

    return decorator


def deprecated_alias(**aliases) -> Callable:
    """
    Used as a decorator when deprecating old function argument names, while
    keeping backwards compatibility. Implementation is inspired from [`StackOverflow`][stack_link].

    [stack_link]: https://stackoverflow.com/questions/49802412/how-to-implement-deprecation-in-python-with-argument-alias

    Examples:
        ```python
        @deprecated_alias(a="alpha", b="beta")
        def simple_sum(alpha, beta):
            return alpha + beta
        ```

    Args:
        **aliases: Dictionary of aliases for a function's arguments.

    Returns:
        Your original function wrapped with the `kwarg` redirection
            function.
    """  # noqa: E501

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            rename_kwargs(func.__name__, kwargs, aliases)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def refactored_function(message: str, category=FutureWarning) -> Callable:
    """Used as a decorator when refactoring functions.

    Implementation is inspired from [`Hacker Noon`][hacker_link].

    [hacker_link]: https://hackernoon.com/why-refactoring-how-to-restructure-python-package-51b89aa91987

    Examples:
        ```python
        @refactored_function(
            message="simple_sum() has been refactored. Use hard_sum() instead."
        )
        def simple_sum(alpha, beta):
            return alpha + beta
        ```

    Args:
        message: Message to use in warning user about refactoring.
        category: Type of Warning. Default is `FutureWarning`.

    Returns:
        Your original function wrapped with the kwarg redirection function.
    """  # noqa: E501

    def decorator(func):
        @wraps(func)
        def emit_warning(*args, **kwargs):
            warn(message, category, stacklevel=find_stack_level())
            return func(*args, **kwargs)

        return emit_warning

    return decorator


def rename_kwargs(func_name: str, kwargs: Dict, aliases: Dict):
    """Used to update deprecated argument names with new names.

    Throws a
    `TypeError` if both arguments are provided, and warns if old alias
    is used. Nothing is returned as the passed `kwargs` are modified
    directly. Implementation is inspired from [`StackOverflow`][stack_link].

    [stack_link]: https://stackoverflow.com/questions/49802412/how-to-implement-deprecation-in-python-with-argument-alias


    Args:
        func_name: name of decorated function.
        kwargs: Arguments supplied to the method.
        aliases: Dictionary of aliases for a function's arguments.

    Raises:
        TypeError: If both arguments are provided.
    """  # noqa: E501
    for old_alias, new_alias in aliases.items():
        if old_alias in kwargs:
            if new_alias in kwargs:
                raise TypeError(
                    f"{func_name} received both {old_alias} and {new_alias}"
                )
            warn(
                f"{old_alias} is deprecated; use {new_alias}",
                DeprecationWarning,
            )
            kwargs[new_alias] = kwargs.pop(old_alias)


def check_column(
    df: pd.DataFrame, column_names: Union[Iterable, str], present: bool = True
):
    """One-liner syntactic sugar for checking the presence or absence
    of columns.

    Examples:
        ```python
        check(df, ["a", "b"], present=True)
        ```

    This will check whether columns `'a'` and `'b'` are present in
    `df`'s columns.

    One can also guarantee that `'a'` and `'b'` are not present
    by switching to `present=False`.

    Args:
        df: The name of the variable.
        column_names: A list of column names we want to check to see if
            present (or absent) in `df`.
        present: If `True` (default), checks to see if all of `column_names`
            are in `df.columns`. If `False`, checks that none of `column_names`
            are in `df.columns`.

    Raises:
        ValueError: If data is not the expected type.
    """
    if isinstance(column_names, str) or not isinstance(column_names, Iterable):
        column_names = [column_names]

    for column_name in column_names:
        if present and column_name not in df.columns:  # skipcq: PYL-R1720
            raise ValueError(f"{column_name} not present in dataframe columns!")
        elif not present and column_name in df.columns:
            raise ValueError(f"{column_name} already present in dataframe columns!")


def skipna(f: Callable) -> Callable:
    """Decorator for escaping `np.nan` and `None` in a function.

    Examples:
        ```python
        df[column].apply(skipna(transform))
        ```

        Can also be used as shown below

        ```python
        @skipna
        def transform(x):
            pass
        ```

    Args:
        f: The function to be wrapped.

    Returns:
        The wrapped function.
    """

    def _wrapped(x, *args, **kwargs):
        if (isinstance(x, float) and np.isnan(x)) or x is None:
            return np.nan
        return f(x, *args, **kwargs)

    return _wrapped


def skiperror(f: Callable, return_x: bool = False, return_val=np.nan) -> Callable:
    """Decorator for escaping any error in a function.

    Examples:
        ```python
        df[column].apply(skiperror(transform, return_val=3, return_x=False))
        ```

        Can also be used as shown below

        ```python
        @skiperror(return_val=3, return_x=False)
        def transform(x):
            pass
        ```

    Args:
        f: The function to be wrapped.
        return_x: Whether or not the original value that caused error
            should be returned.
        return_val: The value to be returned when an error hits.
            Ignored if `return_x` is `True`.

    Returns:
        The wrapped function.
    """

    def _wrapped(x, *args, **kwargs):
        try:
            return f(x, *args, **kwargs)
        except Exception:  # skipcq: PYL-W0703
            if return_x:
                return x
            return return_val

    return _wrapped


def is_connected(url: str) -> bool:
    """This is a helper function to check if the client
    is connected to the internet.

    Examples:

        >>> print(is_connected("www.google.com"))
        True

    Args:
        url: We take a test url to check if we are
            able to create a valid connection.

    Raises:
        OSError: If connection to `URL` cannot be established

    Returns:
        We return a boolean that signifies our connection to the internet
    """
    try:
        sock = socket.create_connection((url, 80))
        if sock is not None:
            sock.close()
            return True
    except OSError as e:
        warn(
            "There was an issue connecting to the internet. "
            "Please see original error below."
        )
        raise e
    return False


def find_stack_level() -> int:
    """Find the first place in the stack that is not inside janitor
    (tests notwithstanding).

    Adapted from Pandas repo.

    Returns:
        Stack level number
    """

    import inspect

    import janitor as jn

    pkg_dir = os.path.abspath(os.path.dirname(jn.__file__))
    test_dir = os.path.join(os.path.dirname(pkg_dir), "tests")

    # https://stackoverflow.com/questions/17407119/python-inspect-stack-is-slow
    frame = inspect.currentframe()
    n = 0
    while frame:
        fname = inspect.getfile(frame)
        if fname.startswith(pkg_dir) and not fname.startswith(test_dir):
            frame = frame.f_back
            n += 1
        else:
            break
    return n


def dynamic_import(file_path: Path):
    """Dynamically import all modules in a directory.

    :param file_path: The path to the file
        containing the modules to import.
    """
    # Iterate through all files in the current directory
    for filename in file_path.glob("*.py"):
        # Check if the file is a Python file and it's not the current __init__.py
        if filename != "__init__.py":
            # Get the module name (without the .py extension)
            module_name = filename.name
            # Dynamically import the module
            module = importlib.import_module(f".{module_name}", package=__name__)
            # Import all symbols from the module into the current namespace
            for name in dir(module):
                if not name.startswith("_"):  # avoid importing private symbols
                    globals()[name] = getattr(module, name)



"""Implementation of `round_to_fraction`"""

from typing import Hashable

import numpy as np
import pandas as pd
import pandas_flavor as pf

from janitor.utils import check, check_column, deprecated_alias


@pf.register_dataframe_method
@deprecated_alias(col_name="column_name")
def round_to_fraction(
    df: pd.DataFrame,
    column_name: Hashable,
    denominator: float,
    digits: float = np.inf,
) -> pd.DataFrame:
    """Round all values in a column to a fraction.

    This method mutates the original DataFrame.

    Taken from [the R package](https://github.com/sfirke/janitor/issues/235).

    Also, optionally round to a specified number of digits.

    Examples:
        Round numeric column to the nearest 1/4 value.

        >>> import numpy as np
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "a1": [1.263, 2.499, np.nan],
        ...         "a2": ["x", "y", "z"],
        ...     }
        ... )
        >>> df
              a1 a2
        0  1.263  x
        1  2.499  y
        2    NaN  z
        >>> df.round_to_fraction("a1", denominator=4)
             a1 a2
        0  1.25  x
        1  2.50  y
        2   NaN  z

    Args:
        df: A pandas DataFrame.
        column_name: Name of column to round to fraction.
        denominator: The denominator of the fraction for rounding. Must be
            a positive number.
        digits: The number of digits for rounding after rounding to the
            fraction. Default is np.inf (i.e. no subsequent rounding).

    Raises:
        ValueError: If `denominator` is not a positive number.

    Returns:
        A pandas DataFrame with a column's values rounded.
    """
    check_column(df, column_name)
    check("denominator", denominator, [float, int])
    check("digits", digits, [float, int])

    if denominator <= 0:
        raise ValueError("denominator is expected to be a positive number.")

    df[column_name] = round(df[column_name] * denominator, 0) / denominator
    if not np.isinf(digits):
        df[column_name] = round(df[column_name], digits)

    return df



"""row_to_names implementation for polars."""

from __future__ import annotations

from functools import singledispatch

from janitor.utils import check, import_message

from .polars_flavor import register_dataframe_method

try:
    import polars as pl
except ImportError:
    import_message(
        submodule="polars",
        package="polars",
        conda_channel="conda-forge",
        pip_install=True,
    )


@register_dataframe_method
def row_to_names(
    df: pl.DataFrame,
    row_numbers: int | list | slice = 0,
    remove_rows: bool = False,
    remove_rows_above: bool = False,
    separator: str = "_",
) -> pl.DataFrame:
    """
    Elevates a row, or rows, to be the column names of a DataFrame.

    Examples:
        Replace column names with the first row.

        >>> import polars as pl
        >>> import janitor.polars
        >>> df = pl.DataFrame(
        ...     {
        ...         "a": ["nums", "6", "9"],
        ...         "b": ["chars", "x", "y"],
        ...     }
        ... )
        >>> df
        shape: (3, 2)
        ┌──────┬───────┐
        │ a    ┆ b     │
        │ ---  ┆ ---   │
        │ str  ┆ str   │
        ╞══════╪═══════╡
        │ nums ┆ chars │
        │ 6    ┆ x     │
        │ 9    ┆ y     │
        └──────┴───────┘
        >>> df.row_to_names(0, remove_rows=True)
        shape: (2, 2)
        ┌──────┬───────┐
        │ nums ┆ chars │
        │ ---  ┆ ---   │
        │ str  ┆ str   │
        ╞══════╪═══════╡
        │ 6    ┆ x     │
        │ 9    ┆ y     │
        └──────┴───────┘
        >>> df.row_to_names(row_numbers=[0, 1], remove_rows=True)
        shape: (1, 2)
        ┌────────┬─────────┐
        │ nums_6 ┆ chars_x │
        │ ---    ┆ ---     │
        │ str    ┆ str     │
        ╞════════╪═════════╡
        │ 9      ┆ y       │
        └────────┴─────────┘

        Remove rows above the elevated row and the elevated row itself.

        >>> df = pl.DataFrame(
        ...     {
        ...         "a": ["bla1", "nums", "6", "9"],
        ...         "b": ["bla2", "chars", "x", "y"],
        ...     }
        ... )
        >>> df
        shape: (4, 2)
        ┌──────┬───────┐
        │ a    ┆ b     │
        │ ---  ┆ ---   │
        │ str  ┆ str   │
        ╞══════╪═══════╡
        │ bla1 ┆ bla2  │
        │ nums ┆ chars │
        │ 6    ┆ x     │
        │ 9    ┆ y     │
        └──────┴───────┘
        >>> df.row_to_names(1, remove_rows=True, remove_rows_above=True)
        shape: (2, 2)
        ┌──────┬───────┐
        │ nums ┆ chars │
        │ ---  ┆ ---   │
        │ str  ┆ str   │
        ╞══════╪═══════╡
        │ 6    ┆ x     │
        │ 9    ┆ y     │
        └──────┴───────┘

    !!! info "New in version 0.28.0"

    Args:
        row_numbers: Position of the row(s) containing the variable names.
            It can be an integer, list or a slice.
        remove_rows: Whether the row(s) should be removed from the DataFrame.
        remove_rows_above: Whether the row(s) above the selected row should
            be removed from the DataFrame.
        separator: Combines the labels into a single string,
            if row_numbers is a list of integers. Default is '_'.

    Returns:
        A polars DataFrame.
    """  # noqa: E501
    return _row_to_names(
        row_numbers,
        df=df,
        remove_rows=remove_rows,
        remove_rows_above=remove_rows_above,
        separator=separator,
    )


@singledispatch
def _row_to_names(
    row_numbers, df, remove_rows, remove_rows_above, separator
) -> pl.DataFrame:
    """
    Base function for row_to_names.
    """
    raise TypeError(
        "row_numbers should be either an integer, "
        "a slice or a list; "
        f"instead got type {type(row_numbers).__name__}"
    )


@_row_to_names.register(int)  # noqa: F811
def _row_to_names_dispatch(  # noqa: F811
    row_numbers, df, remove_rows, remove_rows_above, separator
):
    headers = df.row(row_numbers, named=True)
    headers = {col: str(repl) for col, repl in headers.items()}
    df = df.rename(mapping=headers)
    if remove_rows_above and remove_rows:
        return df.slice(row_numbers + 1)
    elif remove_rows_above:
        return df.slice(row_numbers)
    elif remove_rows:
        expression = pl.int_range(pl.len()).ne(row_numbers)
        return df.filter(expression)
    return df


@_row_to_names.register(slice)  # noqa: F811
def _row_to_names_dispatch(  # noqa: F811
    row_numbers, df, remove_rows, remove_rows_above, separator
):
    if row_numbers.step is not None:
        raise ValueError(
            "The step argument for slice is not supported in row_to_names."
        )
    headers = df.slice(row_numbers.start, row_numbers.stop - row_numbers.start)
    expression = pl.all().str.join(delimiter=separator)
    headers = headers.select(expression).row(0, named=True)
    headers = {col: str(repl) for col, repl in headers.items()}
    df = df.rename(mapping=headers)
    if remove_rows_above and remove_rows:
        return df.slice(row_numbers.stop)
    elif remove_rows_above:
        return df.slice(row_numbers.start)
    elif remove_rows:
        expression = pl.int_range(pl.len()).is_between(
            row_numbers.start, row_numbers.stop, closed="left"
        )
        return df.filter(~expression)
    return df


@_row_to_names.register(list)  # noqa: F811
def _row_to_names_dispatch(  # noqa: F811
    row_numbers, df, remove_rows, remove_rows_above, separator
):
    if remove_rows_above:
        raise ValueError(
            "The remove_rows_above argument is applicable "
            "only if the row_numbers argument is an integer "
            "or a slice."
        )

    for entry in row_numbers:
        check("entry in the row_numbers argument", entry, [int])

    expression = pl.all().gather(row_numbers)
    expression = expression.str.join(delimiter=separator)
    headers = df.select(expression).row(0, named=True)
    headers = {col: str(repl) for col, repl in headers.items()}
    df = df.rename(mapping=headers)
    if remove_rows:
        expression = pl.int_range(pl.len()).is_in(row_numbers)
        return df.filter(~expression)
    return df



"""Implementation of the `jitter` function."""

from typing import Hashable, Iterable, Optional

import numpy as np
import pandas as pd
import pandas_flavor as pf

from janitor.utils import check


@pf.register_dataframe_method
def jitter(
    df: pd.DataFrame,
    column_name: Hashable,
    dest_column_name: str,
    scale: np.number,
    clip: Optional[Iterable[np.number]] = None,
    random_state: Optional[np.number] = None,
) -> pd.DataFrame:
    """Adds Gaussian noise (jitter) to the values of a column.

    A new column will be created containing the values of the original column
    with Gaussian noise added.
    For each value in the column, a Gaussian distribution is created
    having a location (mean) equal to the value
    and a scale (standard deviation) equal to `scale`.
    A random value is then sampled from this distribution,
    which is the jittered value.
    If a tuple is supplied for `clip`,
    then any values of the new column less than `clip[0]`
    will be set to `clip[0]`,
    and any values greater than `clip[1]` will be set to `clip[1]`.
    Additionally, if a numeric value is supplied for `random_state`,
    this value will be used to set the random seed used for sampling.
    NaN values are ignored in this method.

    This method mutates the original DataFrame.

    Examples:
        >>> import numpy as np
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame({"a": [3, 4, 5, np.nan]})
        >>> df
             a
        0  3.0
        1  4.0
        2  5.0
        3  NaN
        >>> df.jitter("a", dest_column_name="a_jit", scale=1, random_state=42)
             a     a_jit
        0  3.0  3.496714
        1  4.0  3.861736
        2  5.0  5.647689
        3  NaN       NaN

    Args:
        df: A pandas DataFrame.
        column_name: Name of the column containing
            values to add Gaussian jitter to.
        dest_column_name: The name of the new column containing the
            jittered values that will be created.
        scale: A positive value multiplied by the original
            column value to determine the scale (standard deviation) of the
            Gaussian distribution to sample from. (A value of zero results in
            no jittering.)
        clip: An iterable of two values (minimum and maximum) to clip
            the jittered values to, default to None.
        random_state: An integer or 1-d array value used to set the random
            seed, default to None.

    Raises:
        TypeError: If `column_name` is not numeric.
        ValueError: If `scale` is not a numerical value
            greater than `0`.
        ValueError: If `clip` is not an iterable of length `2`.
        ValueError: If `clip[0]` is greater than `clip[1]`.

    Returns:
        A pandas DataFrame with a new column containing
            Gaussian-jittered values from another column.
    """

    # Check types
    check("scale", scale, [int, float])

    # Check that `column_name` is a numeric column
    if not np.issubdtype(df[column_name].dtype, np.number):
        raise TypeError(f"{column_name} must be a numeric column.")

    if scale <= 0:
        raise ValueError("`scale` must be a numeric value greater than 0.")
    values = df[column_name]
    if random_state is not None:
        np.random.seed(random_state)
    result = np.random.normal(loc=values, scale=scale)
    if clip:
        # Ensure `clip` has length 2
        if len(clip) != 2:
            raise ValueError("`clip` must be an iterable of length 2.")
        # Ensure the values in `clip` are ordered as min, max
        if clip[1] < clip[0]:
            raise ValueError("`clip[0]` must be less than or equal to `clip[1]`.")
        result = np.clip(result, *clip)
    df[dest_column_name] = result

    return df



"""Functions for cleaning columns/index names and/or column values."""

from __future__ import annotations

import unicodedata

import pandas as pd
import pandas_flavor as pf
from pandas.api.types import is_scalar

from janitor.errors import JanitorError
from janitor.functions.select import get_index_labels
from janitor.functions.utils import _is_str_or_cat
from janitor.utils import deprecated_alias


@pf.register_dataframe_method
@deprecated_alias(preserve_original_columns="preserve_original_labels")
def clean_names(
    df: pd.DataFrame,
    axis: str = "columns",
    column_names: str | list = None,
    strip_underscores: str | bool = None,
    case_type: str = "lower",
    remove_special: bool = False,
    strip_accents: bool = True,
    preserve_original_labels: bool = True,
    enforce_string: bool = True,
    truncate_limit: int = None,
) -> pd.DataFrame:
    """Clean column/index names. It can also be applied to column values.

    Takes all column names, converts them to lowercase,
    then replaces all spaces with underscores.

    By default, column names are converted to string types.
    This can be switched off by passing in `enforce_string=False`.

    This method does not mutate the original DataFrame.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {"Aloha": range(3), "Bell Chart": range(3), "Animals@#$%^": range(3)}
        ... )
        >>> df
           Aloha  Bell Chart  Animals@#$%^
        0      0           0             0
        1      1           1             1
        2      2           2             2
        >>> df.clean_names()
           aloha  bell_chart  animals@#$%^
        0      0           0             0
        1      1           1             1
        2      2           2             2
        >>> df.clean_names(remove_special=True)
           aloha  bell_chart  animals
        0      0           0        0
        1      1           1        1
        2      2           2        2

    !!! summary "Version Changed"

        - 0.26.0
             - Added `axis` and `column_names` parameters.

    Args:
        df: The pandas DataFrame object.
        axis: Whether to clean the labels on the index or columns.
            If `None`, applies to a defined column
            or columns in `column_names`.
        column_names: Clean the values in a column.
            `axis` should be `None`.
            Column selection is possible using the
            [`select`][janitor.functions.select.select] syntax.
        strip_underscores: Removes the outer underscores from all
            column names/values. Default None keeps outer underscores.
            Values can be either 'left', 'right' or 'both'
            or the respective shorthand 'l',
            'r' and True.
        case_type: Whether to make columns lower or uppercase.
            Current case may be preserved with 'preserve',
            while snake case conversion (from CamelCase or camelCase only)
            can be turned on using "snake".
            Default 'lower' makes all characters lowercase.
        remove_special: Remove special characters from columns.
            Only letters, numbers and underscores are preserved.
        strip_accents: Whether or not to remove accents from
            columns names/values.
        preserve_original_labels: Preserve original names.
            This is later retrievable using `df.original_labels`.
            Applies if `axis` is not None.
        enforce_string: Whether or not to convert all
            column names/values to string type.
            Defaults to True, but can be turned off.
            Columns with >1 levels will not be converted by default.
        truncate_limit: Truncates formatted column names/values
            to the specified length.
            Default None does not truncate.

    Raises:
        ValueError: If `axis=None` and `column_names=None`.

    Returns:
        A pandas DataFrame.
    """
    if not axis and not column_names:
        raise ValueError(
            "Kindly provide an argument to `column_names`, if axis is None."
        )
    if axis is None:
        column_names = get_index_labels(arg=column_names, df=df, axis="columns")
        if is_scalar(column_names):
            column_names = [column_names]
        df = df.copy()
        for column_name in column_names:
            df[column_name] = _clean_names(
                obj=df[column_name],
                enforce_string=enforce_string,
                case_type=case_type,
                remove_special=remove_special,
                strip_accents=strip_accents,
                strip_underscores=strip_underscores,
                truncate_limit=truncate_limit,
            )
        return df

    assert axis in {"index", "columns"}
    df = df.copy()
    target_axis = getattr(df, axis)
    if isinstance(target_axis, pd.MultiIndex):
        target_axis = [
            target_axis.get_level_values(number)
            for number in range(target_axis.nlevels)
        ]
        target_axis = [
            _clean_names(
                obj=obj,
                enforce_string=enforce_string,
                case_type=case_type,
                remove_special=remove_special,
                strip_accents=strip_accents,
                strip_underscores=strip_underscores,
                truncate_limit=truncate_limit,
            )
            for obj in target_axis
        ]
        # Convert list of arrays back to MultiIndex
        target_axis = pd.MultiIndex.from_arrays(
            target_axis, names=getattr(df, axis).names
        )
    else:
        target_axis = _clean_names(
            obj=target_axis,
            enforce_string=enforce_string,
            case_type=case_type,
            remove_special=remove_special,
            strip_accents=strip_accents,
            strip_underscores=strip_underscores,
            truncate_limit=truncate_limit,
        )
    # Store the original column names, if enabled by user
    if preserve_original_labels:
        df.__dict__["original_labels"] = getattr(df, axis)
    setattr(df, axis, target_axis)
    return df


def _clean_names(
    obj: pd.Index | pd.Series,
    strip_underscores: str | bool,
    case_type: str,
    remove_special: bool,
    strip_accents: bool,
    enforce_string: bool,
    truncate_limit: int,
) -> pd.Index | pd.Series:
    """
    Generic function to clean labels in a pandas object.
    """
    if enforce_string and not _is_str_or_cat(obj):
        obj = obj.astype(str)
    obj = _change_case(obj=obj, case_type=case_type)
    obj = _normalize_1(obj=obj)
    if remove_special:
        obj = obj.str.replace(pat="[^A-Za-z_\\d]", repl="", regex=True).str.strip()
    if strip_accents:
        obj = _strip_accents(obj=obj)
    obj = obj.str.replace(pat="_+", repl="_", regex=True)
    obj = _strip_underscores_func(
        obj,
        strip_underscores=strip_underscores,
    )
    if truncate_limit:
        obj = obj.str[:truncate_limit]
    return obj


def _change_case(
    obj: pd.Index | pd.Series,
    case_type: str,
) -> pd.Index | pd.Series:
    """Change case of labels in obj."""
    case_types = {"preserve", "upper", "lower", "snake"}
    case_type = case_type.lower()
    if case_type not in case_types:
        raise JanitorError(f"case_type must be one of: {case_types}")

    if case_type == "preserve":
        return obj
    if case_type == "upper":
        return obj.str.upper()
    if case_type == "lower":
        return obj.str.lower()
    # Implementation taken from: https://gist.github.com/jaytaylor/3660565
    # by @jtaylor
    return (
        obj.str.replace(pat=r"(.)([A-Z][a-z]+)", repl=r"\1_\2", regex=True)
        .str.replace(pat=r"([a-z0-9])([A-Z])", repl=r"\1_\2", regex=True)
        .str.lower()
    )


def _normalize_1(obj: pd.Index | pd.Series) -> pd.Index | pd.Series:
    """Perform normalization of labels in obj."""
    FIXES = [
        (r"[ /:,?()\.-]", "_"),
        (r"['’]", ""),
        (r"[\xa0]", "_"),
        # Use capturing groups instead of lookbehind/lookahead
        # for compatibility with pyarrow's RE2 engine (pandas 3.0+).
        (r"(\w)@(\w)", r"\1_\2"),
    ]
    for search, replace in FIXES:
        obj = obj.str.replace(pat=search, repl=replace, regex=True)

    return obj


def _strip_accents(
    obj: pd.Index | pd.Series,
) -> pd.Index | pd.Series:
    """Remove accents from a label.

    Inspired from [StackOverflow][so].

    [so]: https://stackoverflow.com/questions/517923/what-is-the-best-way-to-remove-accents-in-a-python-unicode-strin
    """  # noqa: E501
    return obj.map(
        lambda f: "".join(
            [
                letter
                for letter in unicodedata.normalize("NFD", str(f))
                if not unicodedata.combining(letter)
            ]
        )
    )


def _strip_underscores_func(
    obj: pd.Index | pd.Series,
    strip_underscores: str | bool = None,
) -> pd.Index | pd.Series:
    """Strip underscores."""
    underscore_options = {None, "left", "right", "both", "l", "r", True}
    if strip_underscores not in underscore_options:
        raise JanitorError(f"strip_underscores must be one of: {underscore_options}")
    if strip_underscores in {"left", "l"}:
        return obj.str.lstrip("_")
    if strip_underscores in {"right", "r"}:
        return obj.str.rstrip("_")
    if strip_underscores in {True, "both"}:
        return obj.str.strip("_")
    return obj



from __future__ import annotations

import fnmatch
import inspect
import re
from collections.abc import Callable as dispatch_callable
from dataclasses import dataclass
from functools import singledispatch
from typing import Any, Callable, Literal

import numpy as np
import pandas as pd
import pandas_flavor as pf
from pandas.api.types import (
    is_bool_dtype,
    is_datetime64_dtype,
    is_list_like,
    is_scalar,
)
from pandas.core.common import is_bool_indexer
from pandas.core.groupby.generic import DataFrameGroupBy, SeriesGroupBy

from janitor.functions.utils import _is_str_or_cat
from janitor.utils import check, deprecated_alias, refactored_function


@pf.register_dataframe_groupby_method
@pf.register_dataframe_method
def select_columns(
    df: pd.DataFrame,
    *args: Any,
    invert: bool = False,
) -> pd.DataFrame:
    """Method-chainable selection of columns.

    It accepts a string, shell-like glob strings `(*string*)`,
    regex, slice, array-like object, or a list of the previous options.

    Selection on a MultiIndex on a level, or multiple levels,
    is possible with a dictionary.

    This method does not mutate the original DataFrame.

    Optional ability to invert selection of columns available as well.

    !!!note

        The preferred option when selecting columns or rows in a Pandas DataFrame
        is with `.loc` or `.iloc` methods.
        `select_columns` is primarily for convenience.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> from numpy import nan
        >>> pd.set_option("display.max_columns", None)
        >>> pd.set_option("display.expand_frame_repr", False)
        >>> pd.set_option("max_colwidth", None)
        >>> data = {
        ...     "name": [
        ...         "Cheetah",
        ...         "Owl monkey",
        ...         "Mountain beaver",
        ...         "Greater short-tailed shrew",
        ...         "Cow",
        ...     ],
        ...     "genus": ["Acinonyx", "Aotus", "Aplodontia", "Blarina", "Bos"],
        ...     "vore": ["carni", "omni", "herbi", "omni", "herbi"],
        ...     "order": [
        ...         "Carnivora",
        ...         "Primates",
        ...         "Rodentia",
        ...         "Soricomorpha",
        ...         "Artiodactyla",
        ...     ],
        ...     "conservation": ["lc", nan, "nt", "lc", "domesticated"],
        ...     "sleep_total": [12.1, 17.0, 14.4, 14.9, 4.0],
        ...     "sleep_rem": [nan, 1.8, 2.4, 2.3, 0.7],
        ...     "sleep_cycle": [nan, nan, nan, 0.133333333, 0.666666667],
        ...     "awake": [11.9, 7.0, 9.6, 9.1, 20.0],
        ...     "brainwt": [nan, 0.0155, nan, 0.00029, 0.423],
        ...     "bodywt": [50.0, 0.48, 1.35, 0.019, 600.0],
        ... }
        >>> df = pd.DataFrame(data)
        >>> df
                                 name       genus   vore         order  conservation  sleep_total  sleep_rem  sleep_cycle  awake  brainwt   bodywt
        0                     Cheetah    Acinonyx  carni     Carnivora            lc         12.1        NaN          NaN   11.9      NaN   50.000
        1                  Owl monkey       Aotus   omni      Primates           NaN         17.0        1.8          NaN    7.0  0.01550    0.480
        2             Mountain beaver  Aplodontia  herbi      Rodentia            nt         14.4        2.4          NaN    9.6      NaN    1.350
        3  Greater short-tailed shrew     Blarina   omni  Soricomorpha            lc         14.9        2.3     0.133333    9.1  0.00029    0.019
        4                         Cow         Bos  herbi  Artiodactyla  domesticated          4.0        0.7     0.666667   20.0  0.42300  600.000

        Explicit label selection:
        >>> df.select_columns("name", "order")
                                 name         order
        0                     Cheetah     Carnivora
        1                  Owl monkey      Primates
        2             Mountain beaver      Rodentia
        3  Greater short-tailed shrew  Soricomorpha
        4                         Cow  Artiodactyla

        Selection via globbing:
        >>> df.select_columns("sleep*", "*wt")
           sleep_total  sleep_rem  sleep_cycle  brainwt   bodywt
        0         12.1        NaN          NaN      NaN   50.000
        1         17.0        1.8          NaN  0.01550    0.480
        2         14.4        2.4          NaN      NaN    1.350
        3         14.9        2.3     0.133333  0.00029    0.019
        4          4.0        0.7     0.666667  0.42300  600.000

        Selection via regex:
        >>> import re
        >>> df.select_columns(re.compile(r"o.+er"))
                  order  conservation
        0     Carnivora            lc
        1      Primates           NaN
        2      Rodentia            nt
        3  Soricomorpha            lc
        4  Artiodactyla  domesticated

        Selection via slicing:
        >>> df.select_columns(
        ...     slice("name", "order"), slice("sleep_total", "sleep_cycle")
        ... )
                                 name       genus   vore         order  sleep_total  sleep_rem  sleep_cycle
        0                     Cheetah    Acinonyx  carni     Carnivora         12.1        NaN          NaN
        1                  Owl monkey       Aotus   omni      Primates         17.0        1.8          NaN
        2             Mountain beaver  Aplodontia  herbi      Rodentia         14.4        2.4          NaN
        3  Greater short-tailed shrew     Blarina   omni  Soricomorpha         14.9        2.3     0.133333
        4                         Cow         Bos  herbi  Artiodactyla          4.0        0.7     0.666667

        Selection via callable:
        >>> from pandas.api.types import is_numeric_dtype
        >>> df.select_columns(is_numeric_dtype)
           sleep_total  sleep_rem  sleep_cycle  awake  brainwt   bodywt
        0         12.1        NaN          NaN   11.9      NaN   50.000
        1         17.0        1.8          NaN    7.0  0.01550    0.480
        2         14.4        2.4          NaN    9.6      NaN    1.350
        3         14.9        2.3     0.133333    9.1  0.00029    0.019
        4          4.0        0.7     0.666667   20.0  0.42300  600.000
        >>> df.select_columns(lambda f: f.isna().any())
           conservation  sleep_rem  sleep_cycle  brainwt
        0            lc        NaN          NaN      NaN
        1           NaN        1.8          NaN  0.01550
        2            nt        2.4          NaN      NaN
        3            lc        2.3     0.133333  0.00029
        4  domesticated        0.7     0.666667  0.42300

        Exclude columns with the `invert` parameter:
        >>> df.select_columns(is_numeric_dtype, invert=True)
                                 name       genus   vore         order  conservation
        0                     Cheetah    Acinonyx  carni     Carnivora            lc
        1                  Owl monkey       Aotus   omni      Primates           NaN
        2             Mountain beaver  Aplodontia  herbi      Rodentia            nt
        3  Greater short-tailed shrew     Blarina   omni  Soricomorpha            lc
        4                         Cow         Bos  herbi  Artiodactyla  domesticated

        Exclude columns with the `DropLabel` class:
        >>> from janitor import DropLabel
        >>> df.select_columns(DropLabel(slice("name", "awake")))
           brainwt   bodywt
        0      NaN   50.000
        1  0.01550    0.480
        2      NaN    1.350
        3  0.00029    0.019
        4  0.42300  600.000


        Selection is possible on a grouped object:
        >>> df.groupby("name").select_columns(
        ...     "*wt"
        ... ).min()  # doctest: +NORMALIZE_WHITESPACE
                                    brainwt   bodywt
        name
        Cheetah                         NaN   50.000
        Cow                         0.42300  600.000
        Greater short-tailed shrew  0.00029    0.019
        Mountain beaver                 NaN    1.350
        Owl monkey                  0.01550    0.480


        Selection on MultiIndex columns:
        >>> d = {
        ...     "num_legs": [4, 4, 2, 2],
        ...     "num_wings": [0, 0, 2, 2],
        ...     "class": ["mammal", "mammal", "mammal", "bird"],
        ...     "animal": ["cat", "dog", "bat", "penguin"],
        ...     "locomotion": ["walks", "walks", "flies", "walks"],
        ... }
        >>> df = pd.DataFrame(data=d)
        >>> df = df.set_index(["class", "animal", "locomotion"]).T
        >>> df
        class      mammal                bird
        animal        cat   dog   bat penguin
        locomotion  walks walks flies   walks
        num_legs        4     4     2       2
        num_wings       0     0     2       2

        Selection with a scalar:
        >>> df.select_columns("mammal")  # doctest: +NORMALIZE_WHITESPACE
        class      mammal
        animal        cat   dog   bat
        locomotion  walks walks flies
        num_legs        4     4     2
        num_wings       0     0     2

        Selection with a tuple:
        >>> df.select_columns(("mammal", "bat"))
        class      mammal
        animal        bat
        locomotion  flies
        num_legs        2
        num_wings       2

        Selection within a level is possible with a dictionary,
        where the key is either a level name or number:
        >>> df.select_columns({"animal": "cat"})
        class      mammal
        animal        cat
        locomotion  walks
        num_legs        4
        num_wings       0
        >>> df.select_columns({1: ["bat", "cat"]})  # doctest: +NORMALIZE_WHITESPACE
        class      mammal
        animal        bat   cat
        locomotion  flies walks
        num_legs        2     4
        num_wings       2     0

        Selection on multiple levels:
        >>> df.select_columns({"class": "mammal", "locomotion": "flies"})
        class      mammal
        animal        bat
        locomotion  flies
        num_legs        2
        num_wings       2

        Selection with a regex on a level:
        >>> df.select_columns(
        ...     {"animal": re.compile(".+t$")}
        ... )  # doctest: +NORMALIZE_WHITESPACE
        class      mammal
        animal        cat   bat
        locomotion  walks flies
        num_legs        4     2
        num_wings       0     2

        Selection with a callable on a level:
        >>> df.select_columns(
        ...     {"animal": lambda f: f.str.endswith("t")}
        ... )  # doctest: +NORMALIZE_WHITESPACE
        class      mammal
        animal        cat   bat
        locomotion  walks flies
        num_legs        4     2
        num_wings       0     2

    Args:
        df: A pandas DataFrame, Series or GroupBy object.
        *args: Valid inputs include: an exact column name to look for,
            a shell-style glob string (e.g. `*_thing_*`),
            a regular expression,
            a callable,
            or variable arguments of all the aforementioned.
            A sequence of booleans is also acceptable.
            A dictionary can be used for selection
            on a MultiIndex on different levels.
        invert: Whether or not to invert the selection.
            This will result in the selection
            of the complement of the columns provided.

    Returns:
        A pandas DataFrame, Series, or GroupBy object, with the specified columns selected.
    """  # noqa: E501
    if isinstance(df, DataFrameGroupBy):
        return _get_columns_on_a_grouped_object(group=df, label=list(args))
    return _select(df, columns=list(args), invert=invert)


@pf.register_dataframe_method
@pf.register_series_method
def select_rows(
    df: pd.DataFrame,
    *args: Any,
    invert: bool = False,
) -> pd.DataFrame:
    """Method-chainable selection of rows.

    It accepts a string, shell-like glob strings `(*string*)`,
    regex, slice, array-like object, or a list of the previous options.

    Selection on a MultiIndex on a level, or multiple levels,
    is possible with a dictionary.

    This method does not mutate the original DataFrame.

    Optional ability to invert selection of rows available as well.


    !!! info "New in version 0.24.0"

    !!!note

        The preferred option when selecting columns or rows in a Pandas DataFrame
        is with `.loc` or `.iloc` methods, as they are generally performant.
        `select_rows` is primarily for convenience.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = {"col1": [1, 2], "foo": [3, 4], "col2": [5, 6]}
        >>> df = pd.DataFrame.from_dict(df, orient="index")
        >>> df
              0  1
        col1  1  2
        foo   3  4
        col2  5  6
        >>> df.select_rows("col*")
              0  1
        col1  1  2
        col2  5  6

    More examples can be found in the
    [`select_columns`][janitor.functions.select.select_columns] section.

    Args:
        df: A pandas DataFrame or Series.
        *args: Valid inputs include: an exact index name to look for,
            a shell-style glob string (e.g. `*_thing_*`),
            a regular expression,
            a callable,
            or variable arguments of all the aforementioned.
            A sequence of booleans is also acceptable.
            A dictionary can be used for selection
            on a MultiIndex on different levels.
        invert: Whether or not to invert the selection.
            This will result in the selection
            of the complement of the rows provided.

    Returns:
        A pandas DataFrame or Series with the specified rows selected.
    """  # noqa: E501
    return _select(df, rows=list(args), invert=invert)


@pf.register_dataframe_groupby_method
@pf.register_dataframe_method
@pf.register_series_groupby_method
@pf.register_series_method
@deprecated_alias(rows="index")
@refactored_function(
    message=(
        "This function has been deprecated. "
        "Kindly use `jn.select_columns` or `jn.select_rows` instead."
    )
)
def select(
    df: pd.DataFrame | pd.Series | DataFrameGroupBy,
    *args: tuple,
    index: Any = None,
    columns: Any = None,
    axis: str = "columns",
    invert: bool = False,
) -> pd.DataFrame | pd.Series | DataFrameGroupBy:
    """Method-chainable selection of rows and/or columns.

    It accepts a string, shell-like glob strings `(*string*)`,
    regex, slice, array-like object, or a list of the previous options.

    Selection on a MultiIndex on a level, or multiple levels,
    is possible with a dictionary.

    This method does not mutate the original DataFrame or Series.

    If the pandas object is a Series, selection is possible only on the index.

    Selection can be inverted with the `DropLabel` class.

    Optional ability to invert selection of index/columns available as well.


    !!! info "New in version 0.24.0"


    !!!note

        The preferred option when selecting columns or rows in a Pandas DataFrame
        is with `.loc` or `.iloc` methods, as they are generally performant.
        `select` is primarily for convenience.

    !!!note

        This function has been deprecated.
        Kindly use `jn.select_columns` or `jn.select_rows`

    !!! abstract "Version Changed"

        - 0.26.0
            - Added variable `args`, `invert` and `axis` parameters.
            - `rows` keyword deprecated in favour of `index`.
        - 0.31.0
            - Add support for pd.Series.
        - 0.32.0
            - Add support for DataFrameGroupBy.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     [[1, 2], [4, 5], [7, 8]],
        ...     index=["cobra", "viper", "sidewinder"],
        ...     columns=["max_speed", "shield"],
        ... )
        >>> df
                    max_speed  shield
        cobra               1       2
        viper               4       5
        sidewinder          7       8
        >>> df.select(index="cobra", columns="shield")
               shield
        cobra       2

        Labels can be dropped with the `DropLabel` class:

        >>> df.select(index=DropLabel("cobra"))
                    max_speed  shield
        viper               4       5
        sidewinder          7       8

    More examples can be found in the
    [`select_columns`][janitor.functions.select.select_columns] section.

    Args:
        df: A pandas DataFrame or a GroupBy object.
        *args: Valid inputs include: an exact index name to look for,
            a shell-style glob string (e.g. `*_thing_*`),
            a regular expression,
            a callable,
            or variable arguments of all the aforementioned.
            A sequence of booleans is also acceptable.
            A dictionary can be used for selection
            on a MultiIndex on different levels.
        index: Valid inputs include: an exact label to look for,
            a shell-style glob string (e.g. `*_thing_*`),
            a regular expression,
            a callable,
            or variable arguments of all the aforementioned.
            A sequence of booleans is also acceptable.
            A dictionary can be used for selection
            on a MultiIndex on different levels.
        columns: Valid inputs include: an exact label to look for,
            a shell-style glob string (e.g. `*_thing_*`),
            a regular expression,
            a callable,
            or variable arguments of all the aforementioned.
            A sequence of booleans is also acceptable.
            A dictionary can be used for selection
            on a MultiIndex on different levels.
        invert: Whether or not to invert the selection.
            This will result in the selection
            of the complement of the rows/columns provided.
        axis: Whether the selection should be on the index('index'),
            or columns('columns').
            Applicable only for the variable args parameter.

    Raises:
        ValueError: If args and index/columns are provided.

    Returns:
        A pandas DataFrame or Series with the specified rows and/or columns selected.
    """  # noqa: E501
    if args and isinstance(df, DataFrameGroupBy):
        return _get_columns_on_a_grouped_object(group=df, label=list(args))
    if isinstance(df, DataFrameGroupBy):
        return _get_columns_on_a_grouped_object(group=df, label=[columns])
    if args:
        check("invert", invert, [bool])
        if (index is not None) or (columns is not None):
            raise ValueError(
                "Either provide variable args with the axis parameter, "
                "or provide arguments to the index and/or columns parameters."
            )
        if axis == "index":
            return _select(df, rows=list(args), columns=columns, invert=invert)
        if axis == "columns":
            return _select(df, columns=list(args), rows=index, invert=invert)
        raise ValueError("axis should be either 'index' or 'columns'.")
    return _select(df, rows=index, columns=columns, invert=invert)


def get_index_labels(
    arg: Any, df: pd.DataFrame, axis: Literal["index", "columns"]
) -> pd.Index:
    """Convenience function to get actual labels from column/index

    !!! info "New in version 0.25.0"

    Args:
        arg: Valid inputs include: an exact column name to look for,
            a shell-style glob string (e.g. `*_thing_*`),
            a regular expression,
            a callable,
            or variable arguments of all the aforementioned.
            A sequence of booleans is also acceptable.
            A dictionary can be used for selection
            on a MultiIndex on different levels.
        df: The pandas DataFrame object.
        axis: Should be either `index` or `columns`.

    Returns:
        A pandas Index.
    """
    assert axis in {"index", "columns"}
    index = getattr(df, axis)
    return index[_select_index(arg, df, axis)]


@pf.register_dataframe_groupby_method
def get_columns(group: DataFrameGroupBy | SeriesGroupBy, label: Any) -> pd.DataFrame:
    """
    Get column(s) from a grouped object,
    using the
    [`select`][janitor.functions.select.select] syntax.

    !!! info "New in version 0.25.0"

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     [[1, 2], [4, 5], [7, 8]],
        ...     index=["cobra", "viper", "sidewinder"],
        ...     columns=["max_speed", "shield"],
        ... )
        >>> df
                    max_speed  shield
        cobra               1       2
        viper               4       5
        sidewinder          7       8
        >>> df.groupby(level=0).get_columns("*ed")
                    max_speed
        cobra               1
        viper               4
        sidewinder          7


    Args:
        group: A Pandas GroupBy object.
        label: column(s) to select.

    Returns:
        A pandas DataFrame.
    """
    return _select(group.obj, columns=label, invert=None)


def _get_columns_on_a_grouped_object(
    group: DataFrameGroupBy | SeriesGroupBy, label: Any
) -> DataFrameGroupBy | SeriesGroupBy:
    """
    Helper function for selecting columns on a grouped object,
    using the
    [`select`][janitor.functions.select.select] syntax.

    Args:
        group: A Pandas GroupBy object.
        label: column(s) to select.

    Returns:
        A pandas groupby object.
    """
    check("groupby object", group, [DataFrameGroupBy, SeriesGroupBy])
    label = get_index_labels(label, group.obj, axis="columns")
    label = label if is_scalar(label) else list(label)
    return group[label]


def _select_regex(index, arg, source="regex"):
    "Process regex on a Pandas Index"
    assert source in ("fnmatch", "regex"), source
    try:
        if source == "fnmatch":
            arg, regex = arg
            # fnmatch.translate() produces patterns like (?s:col.*)\Z
            # (or \z on Python 3.12+) which use inline flags (?s:)
            # and \Z/\z anchors not supported by pyarrow's RE2 regex
            # engine (used in pandas 3.0+ with the default
            # string[pyarrow] dtype).
            # Convert to RE2-compatible: strip (?s:...) wrapper and
            # replace \Z/\z with $.
            regex = re.sub(r'^\(\?s:', '', regex)
            regex = re.sub(r'\)\\[zZ]$', '$', regex)
            bools = index.str.match(regex, na=False)
        else:
            bools = index.str.contains(arg, na=False, regex=True)
        if not bools.any():
            raise KeyError(f"No match was returned for '{arg}'")
        return bools
    except Exception as exc:
        raise KeyError(f"No match was returned for '{arg}'") from exc


def _select_callable(arg, func: Callable, axis=None):
    """
    Process a callable on a Pandas DataFrame/Index.
    """
    bools = func(arg)
    bools = np.asanyarray(bools)
    if not is_bool_dtype(bools):
        raise ValueError(
            "The output of the applied callable should be a 1-D boolean array."
        )
    if axis:
        arg = getattr(arg, axis)
    if len(bools) != len(arg):
        raise IndexError(
            f"The boolean array output from the callable {arg} "
            f"has wrong length: "
            f"{len(bools)} instead of {len(arg)}"
        )
    return bools


@dataclass
class DropLabel:
    """Helper class for removing labels within the `select` syntax.

    `label` can be any of the types supported in the `select`,
    `select_rows` and `select_columns` functions.
    An array of integers not matching the labels is returned.

    !!! info "New in version 0.24.0"

    Args:
        label: Label(s) to be dropped from the index.
    """

    label: Any


@singledispatch
def _select_index(arg, df, axis):
    """Base function for selection on a Pandas Index object.

    Returns either an integer, a slice,
    a sequence of booleans, or an array of integers,
    that match the exact location of the target.
    """
    try:
        return getattr(df, axis).get_loc(arg)
    except Exception as exc:
        raise KeyError(f"No match was returned for {arg}") from exc


@_select_index.register(str)  # noqa: F811
def _index_dispatch(arg, df, axis):  # noqa: F811
    """Base function for selection on a Pandas Index object.

    Applies only to strings.
    It is also applicable to shell-like glob strings,
    which are supported by `fnmatch`.

    Returns either a sequence of booleans, an integer,
    or a slice.
    """
    index = getattr(df, axis)
    if isinstance(index, pd.MultiIndex):
        index = index.get_level_values(0)
    if _is_str_or_cat(index) or is_datetime64_dtype(index):
        try:
            return index.get_loc(arg)
        except KeyError as exc:
            if _is_str_or_cat(index):
                if arg == "*":
                    return slice(None)
                # label selection should be case sensitive
                # fix for Github Issue 1160
                # translating to regex solves the case sensitivity
                # and also avoids the list comprehension
                # not that list comprehension is bad - i'd say it is efficient
                # however, the Pandas str.match method used in _select_regex
                # could offer more performance, especially if the
                # underlying array of the index is a PyArrow string array
                return _select_regex(
                    index, (arg, fnmatch.translate(arg)), source="fnmatch"
                )
            raise KeyError(f"No match was returned for '{arg}'") from exc
    raise KeyError(f"No match was returned for '{arg}'")


@_select_index.register(re.Pattern)  # noqa: F811
def _index_dispatch(arg, df, axis):  # noqa: F811
    """Base function for selection on a Pandas Index object.

    Applies only to regular expressions.
    `re.compile` is required for the regular expression.

    Returns an array of booleans.
    """
    index = getattr(df, axis)
    if isinstance(index, pd.MultiIndex):
        index = index.get_level_values(0)
    return _select_regex(index, arg)


@_select_index.register(range)  # noqa: F811
@_select_index.register(slice)  # noqa: F811
def _index_dispatch(arg, df, axis):  # noqa: F811
    """
    Base function for selection on a Pandas Index object.
    Applies only to slices.

    Returns a slice object.
    """
    index = getattr(df, axis)
    if not index.is_monotonic_increasing:
        if not index.is_unique:
            raise ValueError(
                "Non-unique Index labels should be monotonic increasing."
                "Kindly sort the index."
            )
        if is_datetime64_dtype(index):
            raise ValueError(
                "The DatetimeIndex should be monotonic increasing.Kindly sort the index"
            )

    return index._convert_slice_indexer(arg, kind="loc")


@_select_index.register(dispatch_callable)  # noqa: F811
def _index_dispatch(arg, df, axis):  # noqa: F811
    """
    Base function for selection on a Pandas Index object.
    Applies only to callables.

    The callable is applied to the entire DataFrame.

    Returns an array of booleans.
    """
    # special case for selecting dtypes columnwise
    dtypes = (
        arg.__name__
        for _, arg in inspect.getmembers(pd.api.types, inspect.isfunction)
        if arg.__name__.startswith("is") and arg.__name__.endswith("type")
    )
    if (arg.__name__ in dtypes) and (axis == "columns"):
        bools = df.dtypes.map(arg)
        return np.asanyarray(bools)

    return _select_callable(df, arg, axis)


@_select_index.register(dict)  # noqa: F811
def _index_dispatch(arg, df, axis):  # noqa: F811
    """
    Base function for selection on a Pandas Index object.
    Applies only to a dictionary.

    Returns an array of integers.
    """

    index = getattr(df, axis)
    if not isinstance(index, pd.MultiIndex):
        return _select_index(list(arg), df, axis)
    all_str = (isinstance(entry, str) for entry in arg)
    all_str = all(all_str)
    all_int = (isinstance(entry, int) for entry in arg)
    all_int = all(all_int)
    if not all_str | all_int:
        raise TypeError(
            "The keys in the dictionary represent the levels "
            "in the MultiIndex, and should either be all "
            "strings or integers."
        )
    level_label = {}
    for key, value in arg.items():
        if isinstance(value, dispatch_callable):
            indexer = index.get_level_values(key)
            value = _select_callable(indexer, value)
        elif isinstance(value, re.Pattern):
            indexer = index.get_level_values(key)
            value = _select_regex(indexer, value)
        level_label[key] = value

    level_label = {
        index._get_level_number(level): label for level, label in level_label.items()
    }
    level_label = [level_label.get(num, slice(None)) for num in range(index.nlevels)]
    return index.get_locs(level_label)


@_select_index.register(np.ndarray)  # noqa: F811
@_select_index.register(pd.api.extensions.ExtensionArray)  # noqa: F811
@_select_index.register(pd.Index)  # noqa: F811
@_select_index.register(pd.MultiIndex)  # noqa: F811
@_select_index.register(pd.Series)  # noqa: F811
def _index_dispatch(arg, df, axis):  # noqa: F811
    """
    Base function for selection on a Pandas Index object.
    Applies to pd.Series/pd.Index/pd.array/np.ndarray.

    Returns an array of integers.
    """
    index = getattr(df, axis)

    if is_bool_dtype(arg):
        if len(arg) != len(index):
            raise IndexError(
                f"{arg} is a boolean dtype and has wrong length: "
                f"{len(arg)} instead of {len(index)}"
            )
        return np.asanyarray(arg)
    try:
        if isinstance(arg, pd.Series):
            arr = arg.array
        else:
            arr = arg
        if isinstance(index, pd.MultiIndex) and not isinstance(arg, pd.MultiIndex):
            return index.get_locs([arg])
        arr = index.get_indexer_for(arr)
        not_found = arr == -1
        if not_found.all():
            raise KeyError(f"No match was returned for any of the labels in {arg}")
        elif not_found.any():
            not_found = set(arg).difference(index)
            raise KeyError(
                f"No match was returned for these labels in {arg} - {(*not_found,)}"
            )
        return arr
    except Exception as exc:
        raise KeyError(f"No match was returned for {arg}") from exc


@_select_index.register(DropLabel)  # noqa: F811
def _column_sel_dispatch(cols, df, axis):  # noqa: F811
    """
    Base function for selection on a Pandas Index object.
    Returns the inverse of the passed label(s),
    or the set difference if it is part of a list of labels.

    Returns an array of integers.
    """
    arr = _select_index(cols.label, df, axis)
    index = np.arange(getattr(df, axis).size)
    return _index_converter(arr, index)


@_select_index.register(set)
@_select_index.register(list)  # noqa: F811
def _index_dispatch(arg, df, axis):  # noqa: F811
    """
    Base function for selection on a Pandas Index object.
    Applies only to list type.
    It can take any of slice, str, callable, re.Pattern types, ...,
    or a combination of these types.

    Returns an array of integers.
    """
    index = getattr(df, axis)
    if is_bool_indexer(arg):
        if len(arg) != len(index):
            raise ValueError(
                "The length of the list of booleans "
                f"({len(arg)}) does not match "
                f"the length of the DataFrame's {axis}({index.size})."
            )

        return arg

    # shortcut for single unique dtype of scalars
    checks = (is_scalar(entry) for entry in arg)
    if all(checks):
        dtypes = {type(entry) for entry in arg}
        if len(dtypes) == 1:
            indices = index.get_indexer_for(list(arg))
            if (indices != -1).all():
                return indices

    include = []
    exclude = []
    for entry in arg:
        if isinstance(entry, DropLabel):
            exclude.append(entry)
        else:
            outcome = _select_index(entry, df, axis)
            include.append(outcome)
    if exclude:
        if len(exclude) > 1:
            exclude = [entry.label for entry in exclude]
            exclude = DropLabel(exclude)
        else:
            exclude = exclude[0]
        exclude = _select_index(exclude, df, axis)
    len_exclude = len(exclude)
    if len_exclude and not include:
        index_arr = np.arange(getattr(df, axis).size)
        return np.delete(index_arr, exclude)
    if include and len_exclude:
        include = [_index_converter(arr, index) for arr in include]
        include = np.concatenate(include)
        mask = np.isin(include, exclude)
        return include[~mask]
    # single entry does not need to be combined
    # or materialized if possible;
    # this offers more performance
    if len(include) == 1:
        if is_scalar(include[0]):
            return include
        include = include[0]
        if is_list_like(include):
            include = np.asanyarray(include)
        return include
    include = [_index_converter(arr, index) for arr in include]
    return np.concatenate(include)


def _index_converter(arr, index):
    """Converts output from _select_index to an array_like"""
    if is_list_like(arr):
        arr = np.asanyarray(arr)
    if is_bool_dtype(arr):
        arr = arr.nonzero()[0]
    elif isinstance(arr, slice):
        arr = np.arange(index.size)[arr]
    elif isinstance(arr, int):
        arr = np.array([arr])
    return arr


def _select(
    df: pd.DataFrame | pd.Series,
    invert: bool = False,
    rows=None,
    columns=None,
) -> pd.DataFrame | pd.Series:
    """
    Index DataFrame on the index or columns.
    If it is a Series, indexing is only on the index.

    Returns a DataFrame or Series.
    """
    if (columns is not None) and isinstance(df, pd.Series):
        raise ValueError("columns axis is not supported for pd.Series.select")
    if rows is None:
        row_indexer = slice(None)
    else:
        outcome = _select_index([rows], df, axis="index")
        if invert:
            row_indexer = np.ones(df.index.size, dtype=np.bool_)
            row_indexer[outcome] = False
        else:
            row_indexer = outcome
    if isinstance(df, pd.Series):
        return df.iloc[row_indexer]
    if columns is None:
        column_indexer = slice(None)
    else:
        outcome = _select_index([columns], df, axis="columns")
        if invert:
            column_indexer = np.ones(df.columns.size, dtype=np.bool_)
            column_indexer[outcome] = False
        else:
            column_indexer = outcome
    return df.iloc[row_indexer, column_indexer]



# helper functions for !=
import numpy as np
import pandas as pd

from janitor.functions._conditional_join import _binary_search
from janitor.functions._conditional_join._helpers import (
    _convert_array_to_numpy,
    _keep_output,
    _null_checks_cond_join,
    _sort_if_not_monotonic,
)


def _not_equal_indices(
    left: pd.Series,
    right: pd.Series,
    keep: str,
) -> dict | None:
    """
    Use binary search to get indices where
    `left` is exactly  not equal to `right`.

    It is a combination of strictly less than
    and strictly greater than indices.
    """

    dummy = np.array([], dtype=np.intp)

    # deal with nulls
    l1_nulls = dummy
    r1_nulls = dummy
    l2_nulls = dummy
    r2_nulls = dummy
    lt_left = [dummy]
    lt_right = [dummy]
    gt_left = [dummy]
    gt_right = [dummy]
    any_left_nulls = left.isna()
    any_right_nulls = right.isna()
    if any_left_nulls.any():
        l1_nulls = left.index[any_left_nulls.array]
        l1_nulls = l1_nulls.to_numpy(copy=False)
        r1_nulls = right.index
        # avoid NAN duplicates
        if any_right_nulls.any():
            r1_nulls = r1_nulls[~any_right_nulls.array]
        r1_nulls = r1_nulls.to_numpy(copy=False)
        nulls_count = l1_nulls.size
        # blow up nulls to match length of right
        l1_nulls = np.tile(l1_nulls, r1_nulls.size)
        # ensure length of right matches left
        if nulls_count > 1:
            r1_nulls = np.repeat(r1_nulls, nulls_count)
    if any_right_nulls.any():
        r2_nulls = right.index[any_right_nulls.array]
        r2_nulls = r2_nulls.to_numpy(copy=False)
        l2_nulls = left.index
        right = right[~any_right_nulls]
        nulls_count = r2_nulls.size
        # blow up nulls to match length of left
        r2_nulls = np.tile(r2_nulls, l2_nulls.size)
        # ensure length of left matches right
        if nulls_count > 1:
            l2_nulls = np.repeat(l2_nulls, nulls_count)

    l1_nulls = [l1_nulls, l2_nulls]
    r1_nulls = [r1_nulls, r2_nulls]
    check1 = _null_checks_cond_join(series=left)
    check2 = _null_checks_cond_join(series=right)
    if (check1 is None) or (check2 is None):
        lt_left = [dummy]
        lt_right = [dummy]
    else:
        left, _ = check1
        right, _ = check2
        right, _ = _sort_if_not_monotonic(series=right)
        right_index = right.index._values
        left_array = _convert_array_to_numpy(array=left._values)
        right_array = _convert_array_to_numpy(array=right._values)
        outcome = _binary_search._binary_search_lt_first(
            left=left_array, right=right_array, left_index=left.index._values
        )
        if outcome is not None:
            len_right = right.size
            lt_left, search_indices = outcome
            lt_right = [right_index[ind:len_right] for ind in search_indices]
            lt_left = [lt_left.repeat(len_right - search_indices)]
        outcome = _binary_search._binary_search_gt_first(
            left=left_array, right=right_array, left_index=left.index._values
        )
        if outcome is not None:
            gt_left, search_indices = outcome
            gt_right = [right_index[:ind] for ind in search_indices]
            gt_left = [gt_left.repeat(search_indices)]
    lt_left.extend(gt_left)
    lt_left.extend(l1_nulls)
    lt_right.extend(gt_right)
    lt_right.extend(r1_nulls)
    left = np.concatenate(lt_left)
    right = np.concatenate(lt_right)
    if (not left.size) & (not right.size):
        return {
            "left_index": dummy,
            "right_index": dummy,
        }
    outcome = _keep_output(keep, left, right)
    left_index, right_index = outcome
    return {"left_index": left_index, "right_index": right_index}



import janitor_rs
import numpy as np


def _binary_search_lt(
    left: np.ndarray,
    right: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
) -> tuple:
    """
    Get starts for < joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_lt_int64,
        "int32": janitor_rs.binary_search_lt_int32,
        "int16": janitor_rs.binary_search_lt_int16,
        "int8": janitor_rs.binary_search_lt_int8,
        "uint64": janitor_rs.binary_search_lt_uint64,
        "uint32": janitor_rs.binary_search_lt_uint32,
        "uint16": janitor_rs.binary_search_lt_uint16,
        "uint8": janitor_rs.binary_search_lt_uint8,
        "float64": janitor_rs.binary_search_lt_f64,
        "float32": janitor_rs.binary_search_lt_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    return func(left, right, starts, ends)


def _binary_search_le(
    left: np.ndarray,
    right: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
) -> tuple:
    """
    Get starts for <= joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_le_int64,
        "int32": janitor_rs.binary_search_le_int32,
        "int16": janitor_rs.binary_search_le_int16,
        "int8": janitor_rs.binary_search_le_int8,
        "uint64": janitor_rs.binary_search_le_uint64,
        "uint32": janitor_rs.binary_search_le_uint32,
        "uint16": janitor_rs.binary_search_le_uint16,
        "uint8": janitor_rs.binary_search_le_uint8,
        "float64": janitor_rs.binary_search_le_f64,
        "float32": janitor_rs.binary_search_le_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    return func(left, right, starts, ends)


def _binary_search_gt(
    left: np.ndarray,
    right: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
) -> tuple:
    """
    Get ends for > joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_gt_int64,
        "int32": janitor_rs.binary_search_gt_int32,
        "int16": janitor_rs.binary_search_gt_int16,
        "int8": janitor_rs.binary_search_gt_int8,
        "uint64": janitor_rs.binary_search_gt_uint64,
        "uint32": janitor_rs.binary_search_gt_uint32,
        "uint16": janitor_rs.binary_search_gt_uint16,
        "uint8": janitor_rs.binary_search_gt_uint8,
        "float64": janitor_rs.binary_search_gt_f64,
        "float32": janitor_rs.binary_search_gt_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    return func(left, right, starts, ends)


def _binary_search_ge(
    left: np.ndarray,
    right: np.ndarray,
    starts: np.ndarray,
    ends: np.ndarray,
) -> tuple:
    """
    Get ends for >= joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_ge_int64,
        "int32": janitor_rs.binary_search_ge_int32,
        "int16": janitor_rs.binary_search_ge_int16,
        "int8": janitor_rs.binary_search_ge_int8,
        "uint64": janitor_rs.binary_search_ge_uint64,
        "uint32": janitor_rs.binary_search_ge_uint32,
        "uint16": janitor_rs.binary_search_ge_uint16,
        "uint8": janitor_rs.binary_search_ge_uint8,
        "float64": janitor_rs.binary_search_ge_f64,
        "float32": janitor_rs.binary_search_ge_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    return func(left, right, starts, ends)


def _binary_search_lt_first(
    left: np.ndarray,
    right: np.ndarray,
    left_index: np.ndarray,
) -> tuple:
    """
    Get starts for < joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_lt_first_int64,
        "int32": janitor_rs.binary_search_lt_first_int32,
        "int16": janitor_rs.binary_search_lt_first_int16,
        "int8": janitor_rs.binary_search_lt_first_int8,
        "uint64": janitor_rs.binary_search_lt_first_uint64,
        "uint32": janitor_rs.binary_search_lt_first_uint32,
        "uint16": janitor_rs.binary_search_lt_first_uint16,
        "uint8": janitor_rs.binary_search_lt_first_uint8,
        "float64": janitor_rs.binary_search_lt_first_f64,
        "float32": janitor_rs.binary_search_lt_first_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    search_indices, left_index, total = func(left, right, left_index)
    if not total:
        return None
    return left_index, search_indices


def _binary_search_le_first(
    left: np.ndarray, right: np.ndarray, left_index: np.ndarray
) -> tuple:
    """
    Get starts for <= joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_le_first_int64,
        "int32": janitor_rs.binary_search_le_first_int32,
        "int16": janitor_rs.binary_search_le_first_int16,
        "int8": janitor_rs.binary_search_le_first_int8,
        "uint64": janitor_rs.binary_search_le_first_uint64,
        "uint32": janitor_rs.binary_search_le_first_uint32,
        "uint16": janitor_rs.binary_search_le_first_uint16,
        "uint8": janitor_rs.binary_search_le_first_uint8,
        "float64": janitor_rs.binary_search_le_first_f64,
        "float32": janitor_rs.binary_search_le_first_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    search_indices, left_index, total = func(left, right, left_index)
    if not total:
        return None
    return left_index, search_indices


def _binary_search_gt_first(
    left: np.ndarray, right: np.ndarray, left_index: np.ndarray
) -> tuple:
    """
    Get ends for > joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_gt_first_int64,
        "int32": janitor_rs.binary_search_gt_first_int32,
        "int16": janitor_rs.binary_search_gt_first_int16,
        "int8": janitor_rs.binary_search_gt_first_int8,
        "uint64": janitor_rs.binary_search_gt_first_uint64,
        "uint32": janitor_rs.binary_search_gt_first_uint32,
        "uint16": janitor_rs.binary_search_gt_first_uint16,
        "uint8": janitor_rs.binary_search_gt_first_uint8,
        "float64": janitor_rs.binary_search_gt_first_f64,
        "float32": janitor_rs.binary_search_gt_first_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    search_indices, left_index, total = func(left, right, left_index)
    if not total:
        return None
    return left_index, search_indices


def _binary_search_ge_first(
    left: np.ndarray, right: np.ndarray, left_index: np.ndarray
) -> tuple:
    """
    Get ends for >= joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_ge_first_int64,
        "int32": janitor_rs.binary_search_ge_first_int32,
        "int16": janitor_rs.binary_search_ge_first_int16,
        "int8": janitor_rs.binary_search_ge_first_int8,
        "uint64": janitor_rs.binary_search_ge_first_uint64,
        "uint32": janitor_rs.binary_search_ge_first_uint32,
        "uint16": janitor_rs.binary_search_ge_first_uint16,
        "uint8": janitor_rs.binary_search_ge_first_uint8,
        "float64": janitor_rs.binary_search_ge_first_f64,
        "float32": janitor_rs.binary_search_ge_first_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    search_indices, left_index, total = func(left, right, left_index)
    if not total:
        return None
    return left_index, search_indices


def _binary_search_gt_regions(
    left: np.ndarray, right: np.ndarray, left_index: np.ndarray
) -> tuple:
    """
    Get ends for > joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_gt_regions_int64,
        "int32": janitor_rs.binary_search_gt_regions_int32,
        "int16": janitor_rs.binary_search_gt_regions_int16,
        "int8": janitor_rs.binary_search_gt_regions_int8,
        "uint64": janitor_rs.binary_search_gt_regions_uint64,
        "uint32": janitor_rs.binary_search_gt_regions_uint32,
        "uint16": janitor_rs.binary_search_gt_regions_uint16,
        "uint8": janitor_rs.binary_search_gt_regions_uint8,
        "float64": janitor_rs.binary_search_gt_regions_f64,
        "float32": janitor_rs.binary_search_gt_regions_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    search_indices, left_index, total = func(left, right, left_index)
    if not total:
        return None
    return left_index, search_indices


def _binary_search_ge_regions(
    left: np.ndarray, right: np.ndarray, left_index: np.ndarray
) -> tuple:
    """
    Get ends for >= joins
    """
    mapping = {
        "int64": janitor_rs.binary_search_ge_regions_int64,
        "int32": janitor_rs.binary_search_ge_regions_int32,
        "int16": janitor_rs.binary_search_ge_regions_int16,
        "int8": janitor_rs.binary_search_ge_regions_int8,
        "uint64": janitor_rs.binary_search_ge_regions_uint64,
        "uint32": janitor_rs.binary_search_ge_regions_uint32,
        "uint16": janitor_rs.binary_search_ge_regions_uint16,
        "uint8": janitor_rs.binary_search_ge_regions_uint8,
        "float64": janitor_rs.binary_search_ge_regions_f64,
        "float32": janitor_rs.binary_search_ge_regions_f32,
    }
    dtype_name = left.dtype.name
    try:
        func = mapping[dtype_name]
    except KeyError:
        raise KeyError(f"Unsupported data type -> {dtype_name}")
    search_indices, left_index, total = func(left, right, left_index)
    if not total:
        return None
    return left_index, search_indices



"""Implementation of the `tabyl` function for frequency tables."""

from __future__ import annotations

from typing import Hashable

import pandas as pd
import pandas_flavor as pf


@pf.register_dataframe_method
def tabyl(
    df: pd.DataFrame,
    *column_names: Hashable,
    show_na: bool = True,
    show_missing_levels: bool = True,
) -> pd.DataFrame | dict[Hashable, pd.DataFrame]:
    """Create frequency tables (1-way, 2-way, or 3-way).

    This is a Python implementation of R janitor's `tabyl()` function,
    which creates frequency tables as DataFrames. It supports:

    - **1-way tabyl**: Counts and percentages for a single variable
    - **2-way tabyl**: Crosstab (contingency table) of two variables
    - **3-way tabyl**: Dictionary of 2-way tabyls split by a third variable

    This method does not mutate the original DataFrame.

    Examples:
        1-way tabyl (frequency table of a single variable):

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "cyl": [4, 6, 8, 4, 6, 8, 4, 6],
        ...         "gear": [3, 3, 3, 4, 4, 4, 5, 5],
        ...         "am": [0, 0, 0, 1, 1, 1, 1, 1],
        ...     }
        ... )
        >>> df.tabyl("cyl")
           cyl  n  percent
        0    4  3    0.375
        1    6  3    0.375
        2    8  2    0.250

        2-way tabyl (crosstab):

        >>> df.tabyl("cyl", "gear")
           cyl  3  4  5
        0    4  1  1  1
        1    6  1  1  1
        2    8  1  1  0

        3-way tabyl (dictionary of 2-way tabyls split by third variable):

        >>> result = df.tabyl("cyl", "gear", "am")
        >>> result[0]  # Crosstab for am=0
           cyl  3
        0    4  1
        1    6  1
        2    8  1
        >>> result[1]  # Crosstab for am=1
           cyl  4  5
        0    4  1  1
        1    6  1  1
        2    8  1  0

    Args:
        df: The pandas DataFrame object.
        *column_names: One, two, or three column names.
            - 1 column: Returns counts and percentages
            - 2 columns: Returns crosstab (first as rows, second as columns)
            - 3 columns: Returns dict of crosstabs split by third variable
        show_na: Whether to include NA values in the output.
            Defaults to True.
        show_missing_levels: Whether to show zero counts for missing
            categorical levels. Defaults to True.

    Raises:
        ValueError: If fewer than 1 or more than 3 column names are provided.
        KeyError: If any column name is not found in the DataFrame.

    Returns:
        For 1-way and 2-way tabyls: A pandas DataFrame.
        For 3-way tabyls: A dictionary mapping values of the third variable
        to DataFrames (2-way tabyls).
    """
    if len(column_names) == 0:
        raise ValueError("At least one column name must be provided.")
    if len(column_names) > 3:
        raise ValueError("At most three column names can be provided.")

    # Validate that all columns exist
    for col in column_names:
        if col not in df.columns:
            raise KeyError(f"Column '{col}' not found in DataFrame.")

    if len(column_names) == 1:
        return _tabyl_one(
            df,
            column_names[0],
            show_na=show_na,
            show_missing_levels=show_missing_levels,
        )
    elif len(column_names) == 2:
        return _tabyl_two(
            df,
            column_names[0],
            column_names[1],
            show_na=show_na,
            show_missing_levels=show_missing_levels,
        )
    else:
        return _tabyl_three(
            df,
            column_names[0],
            column_names[1],
            column_names[2],
            show_na=show_na,
            show_missing_levels=show_missing_levels,
        )


def _tabyl_one(
    df: pd.DataFrame,
    col: Hashable,
    show_na: bool,
    show_missing_levels: bool,
) -> pd.DataFrame:
    """Create a 1-way frequency table.

    Args:
        df: The pandas DataFrame.
        col: The column name to tabulate.
        show_na: Whether to include NA values.
        show_missing_levels: Whether to show missing categorical levels.

    Returns:
        DataFrame with columns: [col, 'n', 'percent'] and optionally
        'valid_percent' if there are NA values.
    """
    series = df[col]

    # Handle categorical: if not showing missing levels, convert to non-cat
    if not show_missing_levels and hasattr(series, "cat"):
        # Remove unused categories to exclude them from output
        series = series.cat.remove_unused_categories()

    # Handle categorical with missing levels
    if show_missing_levels and hasattr(series, "cat"):
        # For categorical, use all categories
        counts = series.value_counts(dropna=not show_na)
        # Add missing categories with count 0
        all_categories = series.cat.categories
        for cat in all_categories:
            if cat not in counts.index:
                counts[cat] = 0
        counts = counts.sort_index()
    else:
        counts = series.value_counts(dropna=not show_na)
        counts = counts.sort_index()

    total = counts.sum()
    valid_total = series.dropna().shape[0] if show_na else total

    result = pd.DataFrame({col: counts.index, "n": counts.values})

    # Calculate percent (including NA in denominator)
    result["percent"] = result["n"] / total if total > 0 else 0.0

    # Add valid_percent only if there are NA values and show_na is True
    has_na = bool(series.isna().any())
    if has_na and show_na:
        # valid_percent excludes NA from both numerator and denominator
        result["valid_percent"] = result.apply(
            lambda row: (
                row["n"] / valid_total
                if valid_total > 0 and pd.notna(row[col])
                else None
            ),
            axis=1,
        )

    return result.reset_index(drop=True)


def _tabyl_two(
    df: pd.DataFrame,
    row_col: Hashable,
    col_col: Hashable,
    show_na: bool,
    show_missing_levels: bool,
) -> pd.DataFrame:
    """Create a 2-way frequency table (crosstab).

    Args:
        df: The pandas DataFrame.
        row_col: The column for rows.
        col_col: The column for columns.
        show_na: Whether to include NA values.
        show_missing_levels: Whether to show missing categorical levels.

    Returns:
        DataFrame with row_col as first column and counts as remaining columns.
    """
    # Create crosstab
    crosstab = pd.crosstab(
        df[row_col],
        df[col_col],
        dropna=not show_na,
    )

    # Handle missing categorical levels for rows
    if show_missing_levels and hasattr(df[row_col], "cat"):
        all_row_cats = df[row_col].cat.categories
        for cat in all_row_cats:
            if cat not in crosstab.index:
                crosstab.loc[cat] = 0
        crosstab = crosstab.sort_index()

    # Handle missing categorical levels for columns
    if show_missing_levels and hasattr(df[col_col], "cat"):
        all_col_cats = df[col_col].cat.categories
        for cat in all_col_cats:
            if cat not in crosstab.columns:
                crosstab[cat] = 0
        crosstab = crosstab[sorted(crosstab.columns)]

    # Convert to desired format: first column is row_col values
    result = crosstab.reset_index()
    result.columns.name = None  # Remove the column axis name

    return result


def _tabyl_three(
    df: pd.DataFrame,
    row_col: Hashable,
    col_col: Hashable,
    split_col: Hashable,
    show_na: bool,
    show_missing_levels: bool,
) -> dict[Hashable, pd.DataFrame]:
    """Create a 3-way frequency table (dict of crosstabs split by third var).

    Args:
        df: The pandas DataFrame.
        row_col: The column for rows.
        col_col: The column for columns.
        split_col: The column to split by.
        show_na: Whether to include NA values.
        show_missing_levels: Whether to show missing categorical levels.

    Returns:
        Dictionary mapping values of split_col to 2-way tabyl DataFrames.
    """
    result = {}

    # Get unique values of split column
    if show_na:
        split_values = df[split_col].unique()
    else:
        split_values = df[split_col].dropna().unique()

    # Handle missing categorical levels for split column
    if show_missing_levels and hasattr(df[split_col], "cat"):
        all_split_cats = df[split_col].cat.categories
        split_values = list(set(split_values) | set(all_split_cats))

    for val in sorted(split_values, key=lambda x: (pd.isna(x), x)):
        if pd.isna(val):
            mask = df[split_col].isna()
        else:
            mask = df[split_col] == val
        subset = df.loc[mask]

        if len(subset) > 0 or show_missing_levels:
            subset_df = subset if len(subset) > 0 else df.iloc[:0]
            tabyl_result = _tabyl_two(
                subset_df,
                row_col,
                col_col,
                show_na=show_na,
                show_missing_levels=show_missing_levels,
            )
            result[val] = tabyl_result

    return result



import warnings
from enum import Enum
from typing import Any, Hashable, Iterable, Union

import numpy as np
import pandas as pd
import pandas_flavor as pf
from pandas.api.types import is_list_like

from janitor.functions.select import get_index_labels
from janitor.utils import check_column, deprecated_alias, find_stack_level


@pf.register_dataframe_method
@deprecated_alias(columns="column_names")
def encode_categorical(
    df: pd.DataFrame,
    column_names: Union[str, Iterable[str], Hashable] = None,
    **kwargs: Any,
) -> pd.DataFrame:
    """Encode the specified columns with Pandas' [category dtype][cat].

    [cat]: http://pandas.pydata.org/pandas-docs/stable/user_guide/categorical.html

    It is syntactic sugar around `pd.Categorical`.

    This method does not mutate the original DataFrame.

    Simply pass a string, or a sequence of column names to `column_names`;
    alternatively, you can pass kwargs, where the keys are the column names
    and the values can either be None, `sort`, `appearance`
    or a 1-D array-like object.

    - None: column is cast to an unordered categorical.
    - `sort`: column is cast to an ordered categorical,
              with the order defined by the sort-order of the categories.
    - `appearance`: column is cast to an ordered categorical,
                    with the order defined by the order of appearance
                    in the original column.
    - 1d-array-like object: column is cast to an ordered categorical,
                            with the categories and order as specified
                            in the input array.

    `column_names` and `kwargs` parameters cannot be used at the same time.

    Examples:
        Using `column_names`

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "foo": ["b", "b", "a", "c", "b"],
        ...         "bar": range(4, 9),
        ...     }
        ... )
        >>> df
          foo  bar
        0   b    4
        1   b    5
        2   a    6
        3   c    7
        4   b    8
        >>> df.dtypes
        foo    str
        bar     int64
        dtype: object
        >>> enc_df = df.encode_categorical(column_names="foo")
        >>> enc_df.dtypes
        foo    category
        bar       int64
        dtype: object
        >>> enc_df["foo"].cat.categories
        Index(['a', 'b', 'c'], dtype='str')
        >>> enc_df["foo"].cat.ordered
        False

        Using `kwargs` to specify an ordered categorical.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "foo": ["b", "b", "a", "c", "b"],
        ...         "bar": range(4, 9),
        ...     }
        ... )
        >>> df.dtypes
        foo    str
        bar     int64
        dtype: object
        >>> enc_df = df.encode_categorical(foo="appearance")
        >>> enc_df.dtypes
        foo    category
        bar       int64
        dtype: object
        >>> enc_df["foo"].cat.categories
        Index(['b', 'a', 'c'], dtype='str')
        >>> enc_df["foo"].cat.ordered
        True

    Args:
        df: A pandas DataFrame object.
        column_names: A column name or an iterable (list or tuple)
            of column names.
        **kwargs: A mapping from column name to either `None`,
            `'sort'` or `'appearance'`, or a 1-D array. This is useful
            in creating categorical columns that are ordered, or
            if the user needs to explicitly specify the categories.

    Raises:
        ValueError: If both `column_names` and `kwargs` are provided.

    Returns:
        A pandas DataFrame.
    """  # noqa: E501

    if all((column_names, kwargs)):
        raise ValueError("Only one of `column_names` or `kwargs` can be provided.")
    # column_names deal with only category dtype (unordered)
    # kwargs takes care of scenarios where user wants an ordered category
    # or user supplies specific categories to create the categorical
    if column_names is not None:
        column_names = get_index_labels([column_names], df, axis="columns")
        dtypes = {col: "category" for col in column_names}
        return df.astype(dtypes)

    return _computations_as_categorical(df, **kwargs)


def _computations_as_categorical(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    This function handles cases where
    categorical columns are created with an order,
    or specific values supplied for the categories.
    It uses a kwarg, where the key is the column name,
    and the value is either a string, or a 1D array.
    The default for value is None and will return a categorical dtype
    with no order and categories inferred from the column.
    A DataFrame, with categorical columns, is returned.
    """

    categories_dict = _as_categorical_checks(df, **kwargs)

    categories_dtypes = {}

    for column_name, value in categories_dict.items():
        if value is None:
            cat_dtype = pd.CategoricalDtype()
        elif isinstance(value, str):
            if value == _CategoryOrder.SORT.value:
                _, cat_dtype = df[column_name].factorize(sort=True)
            else:
                _, cat_dtype = df[column_name].factorize(sort=False)
            if cat_dtype.empty:
                raise ValueError(
                    "Kindly ensure there is at least "
                    f"one non-null value in {column_name}."
                )
            cat_dtype = pd.CategoricalDtype(categories=cat_dtype, ordered=True)

        else:  # 1-D array
            cat_dtype = pd.CategoricalDtype(categories=value, ordered=True)

        categories_dtypes[column_name] = cat_dtype

    return df.astype(categories_dtypes)


def _as_categorical_checks(df: pd.DataFrame, **kwargs) -> dict:
    """This function raises errors if columns in `kwargs` are
    absent from the dataframe's columns.

    It also raises errors if the value in `kwargs`
    is not a string (`'appearance'` or `'sort'`), or a 1D array.

    This function is executed before proceeding to the computation phase.

    If all checks pass, a dictionary of column names and value is returned.

    Args:
        df: The pandas DataFrame object.
        **kwargs: A pairing of column name and value.

    Raises:
        TypeError: If `value` is not a 1-D array, or a string.
        ValueError: If `value` is a 1-D array, and contains nulls,
            or is non-unique.

    Returns:
        A dictionary.
    """

    check_column(df, kwargs)

    categories_dict = {}

    for column_name, value in kwargs.items():
        # type check
        if (value is not None) and not (is_list_like(value) or isinstance(value, str)):
            raise TypeError(f"{value} should be list-like or a string.")
        if is_list_like(value):
            if not hasattr(value, "shape"):
                value = np.asarray(value)

            if (value.ndim != 1) or isinstance(value, pd.MultiIndex):
                raise ValueError(
                    f"{value} is not a 1-D array. "
                    "Kindly provide a 1-D array-like object."
                )

            if not isinstance(value, (pd.Series, pd.Index)):
                value = pd.Index(value)

            if value.hasnans:
                raise ValueError(
                    "Kindly ensure there are no nulls in the array provided."
                )

            if not value.is_unique:
                raise ValueError(
                    "Kindly provide unique, non-null values for the array provided."
                )

            if value.empty:
                raise ValueError(
                    "Kindly ensure there is at least "
                    "one non-null value in the array provided."
                )

            # uniques, without nulls
            uniques = df[column_name].factorize(sort=False)[-1]
            if uniques.empty:
                raise ValueError(
                    "Kindly ensure there is at least "
                    f"one non-null value in {column_name}."
                )

            missing = uniques.difference(value, sort=False)
            if not missing.empty and (uniques.size > missing.size):
                warnings.warn(
                    f"Values {tuple(missing)} are missing from "
                    f"the provided categories {value} "
                    f"for {column_name}; this may create nulls "
                    "in the new categorical column.",
                    UserWarning,
                    stacklevel=find_stack_level(),
                )

            elif uniques.equals(missing):
                warnings.warn(
                    f"None of the values in {column_name} are in "
                    f"{value}; this might create nulls for all values "
                    f"in the new categorical column.",
                    UserWarning,
                    stacklevel=find_stack_level(),
                )

        elif isinstance(value, str):
            category_order_types = {ent.value for ent in _CategoryOrder}
            if value.lower() not in category_order_types:
                raise ValueError("Argument should be one of 'appearance' or 'sort'.")

        categories_dict[column_name] = value

    return categories_dict


class _CategoryOrder(Enum):
    """
    order types for encode_categorical.
    """

    SORT = "sort"
    APPEARANCE = "appearance"



"""Implementation source for `groupby_agg`."""

from typing import Callable, List, Union

import pandas as pd
import pandas_flavor as pf

from janitor.utils import deprecated_alias, refactored_function


@pf.register_dataframe_method
@deprecated_alias(new_column="new_column_name", agg_column="agg_column_name")
@refactored_function(
    message=(
        "This function will be deprecated in a 1.x release. "
        "Please use `janitor.transform_column` instead."
    )
)
def groupby_agg(
    df: pd.DataFrame,
    by: Union[List, Callable, str],
    new_column_name: str,
    agg_column_name: str,
    agg: Union[Callable, str],
    dropna: bool = True,
) -> pd.DataFrame:
    """Shortcut for assigning a groupby-transform to a new column.

    This method does not mutate the original DataFrame.

    Intended to be the method-chaining equivalent of:

    ```python
    df = df.assign(...=df.groupby(...)[...].transform(...))
    ```

    !!!note

        This function will be deprecated in a 1.x release.
        Please use
        [`jn.transform_column`][janitor.functions.transform_columns.transform_column]
        instead.

    Examples:
        Basic usage.

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "item": ["shoe", "shoe", "bag", "shoe", "bag"],
        ...         "quantity": [100, 120, 75, 200, 25],
        ...     }
        ... )
        >>> df.groupby_agg(
        ...     by="item",
        ...     agg="mean",
        ...     agg_column_name="quantity",
        ...     new_column_name="avg_quantity",
        ... )
           item  quantity  avg_quantity
        0  shoe       100         140.0
        1  shoe       120         140.0
        2   bag        75          50.0
        3  shoe       200         140.0
        4   bag        25          50.0

        Set `dropna=False` to compute the aggregation, treating the null
        values in the `by` column as an isolated "group".

        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "x": ["a", "a", None, "b"],
        ...         "y": [9, 9, 9, 9],
        ...     }
        ... )
        >>> df.groupby_agg(
        ...     by="x",
        ...     agg="count",
        ...     agg_column_name="y",
        ...     new_column_name="y_count",
        ...     dropna=False,
        ... )
              x  y  y_count
        0     a  9        2
        1     a  9        2
        2   NaN  9        1
        3     b  9        1

    Args:
        df: A pandas DataFrame.
        by: Column(s) to groupby on, will be passed into `DataFrame.groupby`.
        new_column_name: Name of the aggregation output column.
        agg_column_name: Name of the column to aggregate over.
        agg: How to aggregate.
        dropna: Whether or not to include null values, if present in the
            `by` column(s). Default is True (null values in `by` are assigned NaN in
            the new column).

    Returns:
        A pandas DataFrame.
    """  # noqa: E501

    return df.assign(
        **{
            new_column_name: df.groupby(by, dropna=dropna)[agg_column_name].transform(
                agg
            ),
        }
    )



"""Implementation for `drop_duplicate_columns`."""

from typing import Hashable

import pandas as pd
import pandas_flavor as pf


@pf.register_dataframe_method
def drop_duplicate_columns(
    df: pd.DataFrame, column_name: Hashable, nth_index: int = 0
) -> pd.DataFrame:
    """Remove a duplicated column specified by `column_name`.

    Specifying `nth_index=0` will remove the first column,
    `nth_index=1` will remove the second column,
    and so on and so forth.

    The corresponding tidyverse R's library is:
    `select(-<column_name>_<nth_index + 1>)`

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "a": range(2, 5),
        ...         "b": range(3, 6),
        ...         "A": range(4, 7),
        ...         "a*": range(6, 9),
        ...     }
        ... ).clean_names(remove_special=True)
        >>> df
           a  b  a  a
        0  2  3  4  6
        1  3  4  5  7
        2  4  5  6  8
        >>> df.drop_duplicate_columns(column_name="a", nth_index=1)
           a  b  a
        0  2  3  6
        1  3  4  7
        2  4  5  8

    Args:
        df: A pandas DataFrame
        column_name: Name of duplicated columns.
        nth_index: Among the duplicated columns,
            select the nth column to drop.

    Returns:
        A pandas DataFrame
    """
    col_indexes = [
        col_idx
        for col_idx, col_name in enumerate(df.columns)
        if col_name == column_name
    ]

    # Select the column to remove based on nth_index.
    removed_col_idx = col_indexes[nth_index]
    # Filter out columns except for the one to be removed.
    filtered_cols = [c_i for c_i, _ in enumerate(df.columns) if c_i != removed_col_idx]

    return df.iloc[:, filtered_cols]



"""Implementation of `shuffle` functions."""

from typing import Any

import pandas as pd
import pandas_flavor as pf


@pf.register_dataframe_method
def shuffle(
    df: pd.DataFrame, random_state: Any = None, reset_index: bool = True
) -> pd.DataFrame:
    """Shuffle the rows of the DataFrame.

    This method does not mutate the original DataFrame.

    Super-sugary syntax! Underneath the hood, we use `df.sample(frac=1)`,
    with the option to set the random state.

    Examples:
        >>> import pandas as pd
        >>> import janitor
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": range(5),
        ...         "col2": list("abcde"),
        ...     }
        ... )
        >>> df
           col1 col2
        0     0    a
        1     1    b
        2     2    c
        3     3    d
        4     4    e
        >>> df.shuffle(random_state=42)
           col1 col2
        0     1    b
        1     4    e
        2     2    c
        3     0    a
        4     3    d

    Args:
        df: A pandas DataFrame.
        random_state: If provided, set a seed for the random number
            generator. Passed to `pd.DataFrame.sample()`.
        reset_index: If True, reset the dataframe index to the default
            RangeIndex.

    Returns:
        A shuffled pandas DataFrame.
    """
    result = df.sample(frac=1, random_state=random_state)
    if reset_index:
        result = result.reset_index(drop=True)
    return result
